# -*- coding: utf-8 -*-
"""
Created on Wed Feb  3 10:33:57 2021

@author: jmmau
"""

from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element

import numpy as np
import json
import svgwrite

class svg():
    
    def __init__(self,input_file):
        ET.register_namespace("","http://www.w3.org/2000/svg")
        ET.register_namespace("inkscape","http://www.inkscape.org/namespaces/inkscape")
        ET.register_namespace("sodipodi","http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd") 
        ET.register_namespace("cc","http://creativecommons.org/ns#") 
        ET.register_namespace("rdf","http://www.w3.org/1999/02/22-rdf-syntax-ns#")      
        self.tree = ET.parse(input_file)
        self.root = self.tree.getroot()
        self.g_list = self.root.findall(".//{http://www.w3.org/2000/svg}g")
        self.N_steps = 1000
        self.input_file = input_file
        self.begin_click = False
        self.begin = ''
        self.anim_id = ''
        self.anim_i = 0
        
        self.V_min_pu = 0.95
        self.V_max_pu = 1.05
        
        self.post_data = {}

    def set_size(self,width,height):
        self.root.attrib['width']  = f'{width}px'
        self.root.attrib['height'] = f'{height}px'
        
    def save(self,output_file=''):
        if output_file=='':
            output_file = f"{self.input_file.replace('.svg','')}_anim.svg"
        self.tree.write(output_file)
        
    def set_text(self,text_id,string):
        for text in self.root.findall('.//{http://www.w3.org/2000/svg}text'):
            if text.attrib['id'] == text_id: 
                text.text = string
        #for tspan in text_obj.findall('.//{http://www.w3.org/2000/svg}tspan'):
        #    tspan.text = string
 
    def set_title(self,text_id,string):
        for text in self.root.findall('.//{http://www.w3.org/2000/svg}text'):
            if text.attrib['id'] == text_id: text_obj = text
        for tspan in text_obj.findall('.//{http://www.w3.org/2000/svg}tspan'):
            tspan.text = string
            
    def set_rect_style(self,object_id,new_style):
        #style="fill:#337ab7"
        for rect in self.root.findall('.//{http://www.w3.org/2000/svg}rect'):
            
            if rect.attrib['id'] == object_id: 
                #print(rect.attrib['style'])
                #if 'style' in rect.attrib:
                #    rect.attrib['style'].update(new_style)
                #else:
                rect.attrib['style'] = new_style

    def set_line_style(self,object_id,new_style):
        #style="fill:#337ab7"
        for rect in self.root.findall('.//{http://www.w3.org/2000/svg}line'):
            if 'id' in rect.attrib:
                if rect.attrib['id'] == object_id: 
                    #if 'style' in rect.attrib:
                    #    rect.attrib['style'].update(new_style)
                    #else:
                    rect.attrib['style'] = new_style

    def set_path_style(self,object_id,new_style):
        #style="fill:#337ab7"
        for path in self.root.findall('.//{http://www.w3.org/2000/svg}path'):
            if 'id' in path.attrib:
                if path.attrib['id'] == object_id: 
                    #if 'style' in rect.attrib:
                    #    rect.attrib['style'].update(new_style)
                    #else:
                    path.attrib['style'] = new_style
                    print(path.attrib['style'])


    def set_color(self,type_,id_,rgb):
        if type_ == 'rect':
            self.set_rect_style(id_,f"fill:#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}")
            
        if type_ == 'line':
            self.set_line_style(id_,f"stroke:#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}")

        if type_ == 'path':
            self.set_path_style(id_,f"stroke:#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}")


    def tostring(self):
        self.svg_str = ET.tostring(self.root).decode()
        return self.svg_str
    
    def set_grid(self,grid,grid_data):
        self.grid=grid
        
        if type(grid_data) == dict:
            data = grid_data

        if type(grid_data) == str:
            data_input = open(grid_data).read().replace("'",'"')
            data = json.loads(data_input)
            
        self.grid_data = data
        
    def set_lines_currents(self):
        
        for line in self.grid_data['lines']:
            #if not 'monitor'in line or not 'vsc_line' in line: continue
            if 'monitor' in line:
                if not line['monitor']: continue
            if 'vsc_line' in line:
                if not line['vsc_line']: continue
                    
            bus_j = line['bus_j']
            bus_k = line['bus_k']

            for ph in ['a','b','c','n']:

                I_max = self.grid_data["line_codes"][line['code']]['I_max']
                i_r = self.grid.get_value(f'i_l_{bus_j}_{bus_k}_{ph}_r') 
                i_i = self.grid.get_value(f'i_l_{bus_j}_{bus_k}_{ph}_i') 
                i = i_r + 1j*i_i
                i_abs = np.abs(i)
                #print(f'l_{bus_j}_{bus_k}_{ph} = {i_abs:8.1f} A')
                if i_abs < 1e-3: continue
                i_sat = np.clip((i_abs/I_max)**2*255,0,255)
                self.set_color('line',f'l_{bus_j}_{bus_k}_{ph}',(int(i_sat),0,0))

    def set_buses_voltages(self):
        
        v_ac_min = 20e6
        min_dc_voltage = 20e6
        min_n_ac_voltage = 20e6
        min_n_dc_voltage = 20e6    
        
        v_ac_max = 0
        max_dc_voltage = 0
        max_n_ac_voltage = 0
        max_n_dc_voltage = 0  


        for bus in self.grid_data['buses']:
            v_r = self.grid.get_value(f'v_{bus["bus"]}_a_r') 
            v_i = self.grid.get_value(f'v_{bus["bus"]}_a_i') 
            v = v_r + 1j*v_i
            v_abs = np.abs(v)
            V_med_pu= 0.5*(self.V_max_pu + self.V_min_pu)
            acdc = 'ac'
            if 'acdc' in bus:
                if bus['acdc'] == 'DC':
                    V_nom = bus['U_kV']*1000
                    acdc = 'dc'
                if bus['acdc'] == 'AC':
                    V_nom = bus['U_kV']/np.sqrt(3)*1000  
                    acdc = 'ac'
            else:
                V_nom = bus['U_kV']/np.sqrt(3)*1000 
                acdc = 'ac'
                
            V_pu = v_abs/V_nom
            
            # when V_pu = V_med_pu color = 0, when V_pu = V_max_pu color = 255 (red)
            # when V_pu = V_med_pu color = 0, when V_pu = V_min_pu color = 255 (blue)
            if V_pu < V_med_pu:
                blue = np.clip(255*((V_pu - V_med_pu)/(self.V_min_pu - V_med_pu))**2,0,255)
                self.set_color('rect',f'{bus["bus"]}',(0,0,int(blue)))  
            if V_pu > V_med_pu:
                red  = np.clip(255*((V_pu - V_med_pu)/(self.V_max_pu - V_med_pu))**2,0,255)
                self.set_color('rect',f'{bus["bus"]}',(int(red),0,0)) 
            
            if acdc == 'ac':
                if V_pu < v_ac_min:
                    self.post_data.update({'v_ac_min':{'bus':bus['bus'],'value':V_pu}})
                    v_ac_min = V_pu
                if V_pu > v_ac_max:
                    self.post_data.update({'v_ac_max':{'bus':bus['bus'],'value':V_pu}})
                    v_ac_max = V_pu                
                
    def set_tooltips(self, output_file):
            
        s = self
        grid = self.grid
        
        for bus in s.grid_data['buses']:
            bus_id = bus['bus']
            bus_elm_list = s.root.findall(f".//*[@id='{bus_id}']")
            if len(bus_elm_list) > 0:
                bus_elm = bus_elm_list[0]
            else:
                print(f'SVG element {bus_id} not found')
                continue
    
            if f"v_{bus['bus']}_n_r" in grid.y_ini_list:
                v_n_r,v_n_i = grid.get_mvalue([f"v_{bus['bus']}_n_r",f"v_{bus['bus']}_n_i"])
                v_n = v_n_r + 1j*v_n_i
            else:
                v_n = 0.0
            for ph,da in zip(['a','b','c'],['data1','data2','data3']):
                v_r,v_i =  grid.get_mvalue([f"v_{bus['bus']}_{ph}_r",f"v_{bus['bus']}_{ph}_i"])
                v = v_r + 1j*v_i
                v_m = np.abs(v-v_n)
                if 'acdc' in bus:
                    v_base = bus['U_kV']*1000
                else:
                    v_base = bus['U_kV']*1000/np.sqrt(3)
                bus_elm.attrib[da] =f"{v_m:20.1f} V / {v_m/v_base:9.2f} pu "
            bus_elm.attrib['data4'] =f"{np.abs(v_n):20.1f} V"   
            bus_elm.attrib['data0'] = 'bus'
    
            p_load_a,q_load_a = -0.001,-0.001
            p_load_b,q_load_b = -0.001,-0.001
            p_load_c,q_load_c = -0.001,-0.001
            if f"p_load_{bus_id}_a" in grid.u_ini_list:
                p_load_a = grid.get_value(f"p_load_{bus_id}_a")
                q_load_a = grid.get_value(f"q_load_{bus_id}_a")
                p_load_b = grid.get_value(f"p_load_{bus_id}_b")
                q_load_b = grid.get_value(f"q_load_{bus_id}_b")
                p_load_c = grid.get_value(f"p_load_{bus_id}_c")
                q_load_c = grid.get_value(f"q_load_{bus_id}_c")
            p_load = p_load_a + p_load_b + p_load_c
            q_load = q_load_a + q_load_b + q_load_c
    
            bus_elm.attrib['data5'] =f"{p_load/1e3:20.1f} kW" 
            bus_elm.attrib['data6'] =f"{q_load/1e3:20.1f} kvar" 
            bus_elm.attrib['data7'] ="xx" 
    
            bus_elm.attrib['class'] = 'tooltip-trigger'
    
        for line in s.grid_data['lines']:
            bus_j = line['bus_j']
            bus_k = line['bus_k']
    
            for ph in ['a','b','c','n']:
    
                line_id = f'l_{bus_j}_{bus_k}_{ph}'
                line_svg =  s.root.findall(f".//*[@id='{line_id}']")
                if len(line_svg)>0:
                    line_elm = line_svg[0]
                else:
                    continue
                line_elm.attrib['data0'] = 'line'
    
                for phi,da in zip(['a','b','c','n'],['data1','data2','data3','data4']):
                    meas_line_id = f'l_{bus_j}_{bus_k}_{phi}'
    
                    i_m = 0.0
                    if f'i_{line_id}_r' in grid.outputs_list:
                        i_r,i_i =  grid.get_mvalue([f"i_{meas_line_id}_r",f"i_{meas_line_id}_i"])
                        i = i_r + 1j*i_i
                        i_m = np.abs(i)      
                    v_base = 1.0
                    line_elm.attrib[da] =f"{i_m:20.1f} A / {i_m/v_base:9.2f} pu "
    
                line_elm.attrib['class'] = 'tooltip-trigger'
    
        for trafo in s.grid_data['transformers']:
            bus_j = trafo['bus_j']
            bus_k = trafo['bus_k']
    
            for wind in [1,2]:
                trafo_id = f'trafo_{bus_j}_{bus_k}_{wind}'
    
                trafo_svg =  s.root.findall(f".//*[@id='{trafo_id}']")
                if len(trafo_svg)>0:
                    trafo_elm = trafo_svg[0]
                else:
                    print(f'No trafo {trafo_id} found')
                    continue
    
                trafo_elm.attrib['data0'] = 'trafo'
    
                string_1 = ''
                for ph in ['a','b','c']:
                    i_r,i_i =  grid.get_mvalue([f'i_t_{bus_j}_{bus_k}_1_{ph}_r',f'i_t_{bus_j}_{bus_k}_1_{ph}_i'])
                    i = i_r + 1j*i_i
                    i_m = np.abs(i)  
                    string_1 += f'{ph}:  {i_m:0.2f}\t'
    
                string_2 = ''
                for ph in ['a','b','c','n']:
                    i_r,i_i =  grid.get_mvalue([f'i_t_{bus_j}_{bus_k}_2_{ph}_r',f'i_t_{bus_j}_{bus_k}_2_{ph}_i'])
                    i = i_r + 1j*i_i
                    i_m = np.abs(i)  
                    string_2 += f'{ph}: {i_m:0.2f}     '
    
                trafo_elm.attrib['data1'] = string_1 + ' A'
                trafo_elm.attrib['data2'] = string_2 + ' A'
                trafo_elm.attrib['data3'] = 'trafo'
                trafo_elm.attrib['data4'] = 'trafo'
            trafo_elm.attrib['class'] = 'tooltip-trigger'
    
        if 'vscs' in self.grid_data:
            vscs_list = self.grid_data['vscs']
        else:
            vscs_list = []
            
        for vsc in vscs_list:

            if not 'bus_ac' in vsc: continue

            bus_ac = vsc['bus_ac']
            bus_dc = vsc['bus_dc']
            vsc_id = f'vsc_{bus_ac}_{bus_dc}'
    
            vsc_svg =  s.root.findall(f".//*[@id='{vsc_id}']")
            if len(vsc_svg)>0:
                vsc_elm = vsc_svg[0]
            else:
                print(f'No VSC {vsc_id} found')
                continue
            vsc_elm.attrib['data0'] = 'vsc'
            p_dc = grid.get_value(f'p_vsc_{bus_dc}')
            vsc_elm.attrib['data1'] = 'VSC'
    
            p_ac = grid.get_value(f'p_vsc_{bus_ac}')   
            p_dc = grid.get_value(f'p_vsc_{bus_dc}')   
            vsc_elm.attrib['data2'] = f'  Pac = {p_ac/1e3:0.1f} kW,  Pdc = {p_dc/1e3:0.1f} kW'
    
            p_loss = grid.get_value(f'p_vsc_loss_{bus_ac}')
            vsc_elm.attrib['data3'] = f'  Losses = {p_loss/1e3:0.1f} kW'
    
            vsc_elm.attrib['class'] = 'tooltip-trigger'
            
        script = '''
        <![CDATA[
        		(function() {
        			var svg = document.getElementById('document_id');
        			var tooltip_1 = svg.getElementById('tooltip1');
                var tooltipText_1_1 = tooltip_1.getElementsByTagName('text')[0];
                var tooltipText_1_2 = tooltip_1.getElementsByTagName('text')[1];
                var tooltipText_1_3 = tooltip_1.getElementsByTagName('text')[2];
                var tooltipText_1_4 = tooltip_1.getElementsByTagName('text')[3];
                var tooltipText_1_5 = tooltip_1.getElementsByTagName('text')[4];
                var tooltipText_1_6 = tooltip_1.getElementsByTagName('text')[5];
        			var tooltip_2 = svg.getElementById('tooltip2');
                var tooltipText_2_1 = tooltip_2.getElementsByTagName('text')[0];
                var tooltipText_2_2 = tooltip_2.getElementsByTagName('text')[1];
                var tooltipText_2_3 = tooltip_2.getElementsByTagName('text')[2];
                var tooltipText_2_4 = tooltip_2.getElementsByTagName('text')[3];
                var tooltipText_2_5 = tooltip_2.getElementsByTagName('text')[4];
                var tooltipText_2_6 = tooltip_2.getElementsByTagName('text')[5];
        			var tooltip_3 = svg.getElementById('tooltip3');
                var tooltipText_3_1 = tooltip_3.getElementsByTagName('text')[0];
                var tooltipText_3_2 = tooltip_3.getElementsByTagName('text')[1];
                var tooltipText_3_3 = tooltip_3.getElementsByTagName('text')[2];
                var tooltipText_3_4 = tooltip_3.getElementsByTagName('text')[3];
        
                var tooltipRects = tooltip_1.getElementsByTagName('rect');
                var triggers = svg.getElementsByClassName('tooltip-trigger');
                for (var i = 0; i < triggers.length; i++) {
                    triggers[i].addEventListener('mousemove', showTooltip);
                    triggers[i].addEventListener('mouseout', hideTooltip);
                }
                function showTooltip(evt) {
                    var CTM = svg.getScreenCTM();
                    var x_unsat =  (evt.clientX - CTM.e) / CTM.a
                    var x = x_unsat;
                    if  (x_unsat>x_max) {var x = x_max}
                    if  (x_unsat<x_min) {var x = x_min}
        
                   
                    var y_unsat = (evt.clientY - CTM.f) / CTM.d + 30;
                    var y = y_unsat;
                    if  (y_unsat>y_max) {var y = y_max - y_height}
                   
                    
                    if (evt.target.getAttributeNS(null, "data0") == "bus") {
                        tooltip_1.setAttributeNS(null, "transform", "translate(" + x + " " + y + ")");
                        tooltip_1.setAttributeNS(null, "visibility", "visible")}
                    
                    if (evt.target.getAttributeNS(null, "data0") == "line") {
                        tooltip_2.setAttributeNS(null, "transform", "translate(" + x + " " + y + ")");
                        tooltip_2.setAttributeNS(null, "visibility", "visible")}
        
                    if (evt.target.getAttributeNS(null, "data0") == "vsc") {
                        tooltip_3.setAttributeNS(null, "transform", "translate(" + x + " " + y + ")");
                        tooltip_3.setAttributeNS(null, "visibility", "visible")}
                        
                    if (evt.target.getAttributeNS(null, "data0") == "trafo") {
                        tooltip_3.setAttributeNS(null, "transform", "translate(" + x + " " + y + ")");
                        tooltip_3.setAttributeNS(null, "visibility", "visible")}    
                        
                    tooltipText_1_1.firstChild.data = evt.target.getAttributeNS(null, "data1");
                    tooltipText_1_2.firstChild.data = evt.target.getAttributeNS(null, "data2");
                    tooltipText_1_3.firstChild.data = evt.target.getAttributeNS(null, "data3");
                    tooltipText_1_4.firstChild.data = evt.target.getAttributeNS(null, "data4");
                    tooltipText_1_5.firstChild.data = evt.target.getAttributeNS(null, "data5");
                    tooltipText_1_6.firstChild.data = evt.target.getAttributeNS(null, "data6");
         
                    tooltipText_2_1.firstChild.data = evt.target.getAttributeNS(null, "data1");
                    tooltipText_2_2.firstChild.data = evt.target.getAttributeNS(null, "data2");
                    tooltipText_2_3.firstChild.data = evt.target.getAttributeNS(null, "data3");
                    tooltipText_2_4.firstChild.data = evt.target.getAttributeNS(null, "data4");
                    tooltipText_2_5.firstChild.data = evt.target.getAttributeNS(null, "data5");
                    tooltipText_2_6.firstChild.data = evt.target.getAttributeNS(null, "data6");
        
                    tooltipText_3_1.firstChild.data = evt.target.getAttributeNS(null, "data1");
                    tooltipText_3_2.firstChild.data = evt.target.getAttributeNS(null, "data2");
                    tooltipText_3_3.firstChild.data = evt.target.getAttributeNS(null, "data3");
                    tooltipText_3_4.firstChild.data = evt.target.getAttributeNS(null, "data4");
                    
        				var length = tooltipText_1_1.getComputedTextLength()+200;
        				for (var i = 0; i < tooltipRects.length; i++) {
        					tooltipRects[i].setAttributeNS(null, "width", length + 8);
        				}
        			}
        			function hideTooltip(evt) {
        				tooltip_1.setAttributeNS(null, "visibility", "hidden");
        				tooltip_2.setAttributeNS(null, "visibility", "hidden");
        				tooltip_3.setAttributeNS(null, "visibility", "hidden");
        			}
        		})()
            ]]>'''
        
        
        
        element = Element('script')
        element.attrib['type'] ="text/ecmascript"
        element.attrib['id'] = "script15"
        element.text = 'scriptplace'
        self.root.append(element)
        
        self.set_lines_currents()
        self.set_buses_voltages()
            
        out = self.tostring().replace('scriptplace',script)
        width = float(self.root.attrib['width'])
        height = float(self.root.attrib['height'])
        document_id = self.root.attrib['id']
        
        tooltip_width = 320
        tooltip_height = 100
        
        out=out.replace('x_max',f'{width-tooltip_width/2}' )
        out=out.replace('x_min',f'{tooltip_width/2}' )
        out=out.replace('y_max',f'{height-tooltip_height}' )
        out=out.replace('y_height',f'{tooltip_height}' )
        out=out.replace('document_id', document_id)
        
        with open(output_file,'w') as fobj:
            fobj.write(out)  
            
            
class animatesvg():
    
    def __init__(self,input_file,group_id):
        ET.register_namespace("","http://www.w3.org/2000/svg")
        ET.register_namespace("inkscape","http://www.inkscape.org/namespaces/inkscape")
        ET.register_namespace("sodipodi","http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd") 
        ET.register_namespace("cc","http://creativecommons.org/ns#") 
        ET.register_namespace("rdf","http://www.w3.org/1999/02/22-rdf-syntax-ns#")      
        self.tree = ET.parse(input_file)
        self.root = self.tree.getroot()
        self.g_list = self.root.findall(".//{http://www.w3.org/2000/svg}g")
        self.group_id = group_id
        self.N_steps = 1000
        self.input_file = input_file
        self.begin_click = False
        self.begin = ''
        self.anim_id = ''
        self.anim_i = 0
 
    def set_size(self,width,height):
        self.root.attrib['width']  = f'{width}px'
        self.root.attrib['height'] = f'{height}px'
        
    def rotate(self,times,angle,x,y):
        if type(x) == float:
            x = times*0+x
        if type(y) == float:
            y = times*0+y            
        t_end = times[-1]
        N_t = len(times)   
        keyTimes = ""
        keyPoints = ""
        for it in range(N_t):
            keyTimes  += f'{times[it]/t_end};'
            keyPoints += f'{angle[it]},{x[it]},{y[it]};'         
        keyTimes  = keyTimes[:-1].replace("'",'"')    
        keyPoints = keyPoints[:-1].replace("'",'"') 
        
        for item in self.g_list:
            if item.attrib['id'] == self.group_id:

                anim = ET.Element("animateTransform")
                if self.anim_id != '':
                    anim.set('id',self.anim_id)
                anim.set('calcMode',"linear")
                anim.set('additive',"sum")
                anim.set('attributeType',"xml")
                anim.set('attributeName',"transform")
                anim.set('type',"rotate")
                anim.set('dur',f"{t_end}s")
                anim.set('fill',"freeze") 
                #anim.set('repeatCount',"indefinite") 
                if self.begin != '':
                    anim.set('begin',self.begin)
                if self.begin_click:
                    anim.set('begin',"click")
                anim.set('values',f"{keyPoints}")
                anim.set('keyTimes',f"{keyTimes}")  
                item.insert(0, anim)
                
    def translate(self,times,x,y):  
        t_end = times[-1]
        N_t = len(times)   
        keyTimes = ""
        keyPoints = ""
        for it in range(N_t):
            keyTimes  += f'{times[it]/t_end:0.5f};'
            keyPoints += f'{float(x[it]):0.3f},{float(y[it]):0.3f};'         
        keyTimes  = keyTimes[:-1].replace("'",'"')    
        keyPoints = keyPoints[:-1].replace("'",'"') 

        for item in self.g_list:
            if item.attrib['id'] == self.group_id:
                anim = ET.Element("animateTransform")
                if self.anim_id != '':
                    anim.set('id',self.anim_id)
                anim.set('calcMode',"linear")
                anim.set('additive',"sum")
                anim.set('attributeType',"xml")
                anim.set('attributeName',"transform")
                anim.set('type',"translate")
                anim.set('dur',f"{t_end}s")
                anim.set('fill',"freeze") 
                #anim.set('repeatCount',"indefinite") 
                if self.begin != '':
                    anim.set('begin',self.begin)
                if self.begin_click:
                    anim.set('begin',"click")
                anim.set('values',f"{keyPoints}")
                anim.set('keyTimes',f"{keyTimes}")  
                item.insert(0, anim)

    def scale(self,times,x_0,y_0,sx,sy):  
        t_end = times[-1]
        N_t = len(times)   
        keyTimes = ""
        keyPoints = ""
        for it in range(N_t):
            keyTimes  += f'{times[it]/t_end:0.5f};'
            keyPoints += f'{sx[it]:0.3f},{sy[it]:0.3f};'         
        keyTimes  = keyTimes[:-1].replace("'",'"')    
        keyPoints = keyPoints[:-1].replace("'",'"') 

        for item in self.g_list:
            if item.attrib['id'] == self.group_id:
                anim = ET.Element("animateTransform")
                if self.anim_id != '':
                    anim.set('id',self.anim_id)
                
                anim.set('calcMode',"linear")
                anim.set('additive',"sum")
                anim.set('attributeType',"xml")
                anim.set('attributeName',"transform")
                anim.set('type',"scale")
                anim.set('dur',f"{t_end}s")
                anim.set('fill',"freeze") 
                #anim.set('repeatCount',"indefinite") 
                if self.begin != '':
                    anim.set('begin',self.begin)
                if self.begin_click:
                    anim.set('begin',"click")
                anim.set('values',f"{keyPoints}")
                anim.set('keyTimes',f"{keyTimes}")  
                item.insert(0, anim)
        x = times*0+x_0*(1-sx)
        y = times*0+y_0*(1-sy)
        self.translate(times,x,y)


    def save(self,output_file=''):
        if output_file=='':
            output_file = f"{self.input_file.replace('.svg','')}_anim.svg"
        self.tree.write(output_file)
        
    def reduce(self,times,values):
        N_t = len(times)
        mask_values = values
        
        mask = np.array([True]*N_t)
        increment = np.abs((mask_values.max()-mask_values.min())/self.N_steps)
        mask[0:(N_t-1)] =  np.abs(np.diff(mask_values))>increment
        mask[0] = True
        mask[-1] = True
        
        self.mask = mask
        self.increment = increment
        return self.mask
    
    
def grid2svg(input_json,output_svg):
    '''
    

    Parameters
    ----------
    input_json : if string name of the file containing the grid data.
                 if dict, a dictionary with the grid data.
    
    output_svg : string
        path for the output SVG file.

    Returns
    -------
    None.

    '''
    

    if type(input_json) == dict:
        data = input_json
        
    if type(input_json) == str:
        json_data = open(input_json).read().replace("'",'"')
        data = json.loads(json_data)
                
                
                
    

    dwg = svgwrite.Drawing(output_svg, profile='full', size=(640, 800))

    buses = data['buses']
    lines = data['lines']
    trafos = data['transformers']

    offset_x = 100
    offset_y = 50
    scale_x = 1.5
    scale_y = 1.5

    bus2posxy = {}
    for bus in buses:
        pos_x = (bus['pos_x']+offset_x)*scale_x
        pos_y = (offset_y-bus['pos_y'])*scale_y
        bus2posxy.update({bus['bus']:{'pos_x':pos_x,'pos_y':pos_y}})

        dwg.add(svgwrite.shapes.Rect(insert=(pos_x-2.5, pos_y-2.5), size=(15, 15),id = bus['bus']))
        #dwg.add(dwg.text(bus['bus'], insert=(pos_x+3, pos_y-3), fill='black'))
        dwg.add(dwg.text(bus['bus'], insert=(pos_x+5, pos_y-5), fill='black',id = f"{bus['bus']}_tt")) #,visibility='hidden'))

        #a = svgwrite.animate.Set(href=f"{bus['bus']}_tt", attributeName="visibility",
                             #begin="hidden",
        #                     to="visible",
        #                     begin=f"{bus['bus']}.mouseover",
        #                     end=f"{bus['bus']}.mouseout")

        #dwg.add(a)

    for line in lines:

        x1 = bus2posxy[line['bus_j']]['pos_x']
        y1 = bus2posxy[line['bus_j']]['pos_y']
        x2 = bus2posxy[line['bus_k']]['pos_x']
        y2 = bus2posxy[line['bus_k']]['pos_y']
        bus_j = line['bus_j']
        bus_k = line['bus_k']

        v = 0.0
        h = 0.0

        vertical = True
        Dv = 0.0
        Dh = 3.0
        if np.abs(x2 - x1) < np.abs(y2 - y1):
            Dv = 3.0
            Dh = 0.0
            vertical = False

        for ph in ['a','b','c','n']:
            dwg.add(dwg.line(start=(x1+v, y1+h), end=(x2+v,y2+h), 
                             stroke=svgwrite.rgb(10, 10, 16, '%'), id = f'l_{bus_j}_{bus_k}_{ph}',
                             stroke_width=2))
            v += Dv
            h += Dh

    for trafo in trafos:

        x1 = bus2posxy[trafo['bus_j']]['pos_x']
        y1 = bus2posxy[trafo['bus_j']]['pos_y']
        x2 = bus2posxy[trafo['bus_k']]['pos_x']
        y2 = bus2posxy[trafo['bus_k']]['pos_y']

        dwg.add(dwg.line(start=(x1, y1), end=(x2,y2), stroke=svgwrite.rgb(10, 10, 16, '%')))


    dwg.save()
    
    
def results2svg(grid,data_input,svg_input,svg_output):
    
    
    if type(data_input) == dict:
        data = data_input
        
    if type(data_input) == str:
        data_input = open(data_input).read().replace("'",'"')
        data = json.loads(data_input)

    svg_new = svg(svg_input)

    buses = data['buses']
    lines = data['lines']

    trafos = data['transformers']

    for line in lines:
        #if not 'monitor'in line or not 'vsc_line' in line: continue
        if 'monitor' in line:
            if not line['monitor']: continue
        if 'vsc_line' in line:
            if not line['vsc_line']: continue
        bus_j = line['bus_j']
        bus_k = line['bus_k']

        for ph in ['a','b','c','n']:

            I_max = data["line_codes"][line['code']]['I_max']
            i_r = grid.get_value(f'i_l_{bus_j}_{bus_k}_{ph}_r') 
            i_i = grid.get_value(f'i_l_{bus_j}_{bus_k}_{ph}_i') 
            i = i_r + 1j*i_i
            i_abs = np.abs(i)
            #print(f'l_{bus_j}_{bus_k}_{ph} = {i_abs:8.1f} A')
            if i_abs < 1e-3: continue
            i_sat = np.clip(i_abs**2/I_max**2*255,0,255)
            svg_new.set_color('line',f'l_{bus_j}_{bus_k}_{ph}',(int(i_sat),0,0))

    for bus in buses:
        #if not 'monitor'in line or not 'vsc_line' in line: continue

        
        U_nom  = bus['U_kV']
        v_r = grid.get_value(f'v_{bus["bus"]}_a_r') 
        v_i = grid.get_value(f'v_{bus["bus"]}_a_i') 
        v = v_r + 1j*v_i
        v_abs = np.abs(v)
        #print(f'l_{bus_j}_{bus_k}_{ph} = {i_abs:8.1f} A')
        if 'acdc' in bus:
            if bus['acdc'] == 'DC':
                red  = np.clip(((v_abs/(1e3*U_nom))**2-1)*255,0,255)
                blue = np.clip(3*(1-(v_abs/(1e3*U_nom))**2)*255,0,255)
                svg_new.set_color('rect',f'{bus["bus"]}',(int(red),0,int(blue)))  
            if bus['acdc'] == 'AC':
                red  = np.clip(((np.sqrt(3)*v_abs/(1e3*U_nom))**2-1)*255,0,255)
                blue = np.clip(3*(1-(np.sqrt(3)*v_abs/(1e3*U_nom))**2)*255,0,255)
                svg_new.set_color('rect',f'{bus["bus"]}',(int(red),0,int(blue)))  
        else:
            red  = np.clip(((np.sqrt(3)*v_abs/(1e3*U_nom))**2-1)*255,0,255)
            blue = np.clip(3*(1-(np.sqrt(3)*v_abs/(1e3*U_nom))**2)*255,0,255)
            svg_new.set_color('rect',f'{bus["bus"]}',(int(red),0,int(blue)))

    svg_new.save(svg_output)
    
    return svg_new


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020

@author: jmmauricio
"""

import numpy as np
from pydae.tools import get_v,get_i,get_s
import json
from collections import namedtuple
import numba 

class grid(object):
    
    def __init__(self,syst):
    #def bokeh_tools(data):

        self.syst = syst

        self.s_radio_scale = 0.01
        self.s_radio_max = 20
        self.s_radio_min = 1
        
        with np.load('matrices.npz') as data:
            Y_primitive = data['Y_primitive']
            A_conect = data['A_conect']
            nodes_list = data['nodes_list']
            node_sorter = data['node_sorter']
            Y_vv = data['Y_vv']
            Y_vi = data['Y_vi']        
            N_v = int(data['N_v']) 

            
        self.nodes_list = nodes_list
        self.Y_primitive = Y_primitive
        self.A_conect = A_conect
        self.node_sorter = node_sorter
        self.Y_vv = Y_vv
        self.Y_vi = Y_vi
        self.N_v = N_v

        json_file = 'grid_data.json'
        json_file = json_file
        json_data = open(json_file).read().replace("'",'"')
        data = json.loads(json_data)
        
        self.buses = data['buses']
        if 'transformers' in  data:
            self.transformers = data['transformers']
        else:
            self.transformers = []

        self.lines = data['lines']
        self.loads = data['loads']
        
        if 'vscs' in data:
            self.vscs = data['vscs']
        else: self.vscs = []

        
    def dae2vi(self):
        '''
        For obtaining line currents from node voltages after power flow is solved.

        Returns
        -------
        None.

        '''
        n2a = {'1':'a','2':'b','3':'c','4':'n'}
        a2n = {'a':1,'b':2,'c':3,'n':4}
        
        V_node_list = []
        I_node_list = [0.0]*len(self.nodes_list)
        self.I_node_list = I_node_list
        for item in self.nodes_list:
            bus_name,phase_name = item.split('.')
            #i = get_i(self.syst,bus_name,phase_name=n2a[phase_name],i_type='phasor',dq_name='ri')
            #I_node_list += [i]
            v = get_v(self.syst,bus_name,phase_name=n2a[phase_name],v_type='phasor',dq_name='ri')
            V_node_list += [v]


 
        V_node = np.array(V_node_list).reshape(len(V_node_list),1)
        
        V_known = np.copy(V_node[:self.N_v])
        V_unknown = np.copy(V_node[self.N_v:])
        I_unknown = self.Y_vv @ V_known + self.Y_vi @ V_unknown
        #self.I_node = I_node
        self.V_node = V_node
        self.I_unknown = I_unknown
        self.I_known = np.array(I_node_list).reshape(len(I_node_list),1)

        self.I_node = np.vstack((self.I_unknown,self.I_known))
 
        for load in self.loads:
            bus_name = load['bus']
            if load['type'] == '3P+N':
                for ph in ['a','b','c','n']:
                    idx = list(self.nodes_list).index(f"{load['bus']}.{a2n[ph]}") 
                    i_ = get_i(self.syst,'load_' + bus_name,phase_name=ph,i_type='phasor',dq_name='ri')
                    self.I_node[idx] += i_ 

            if load['type'] == '1P+N':
                ph = load['bus_nodes'][0]
                idx = list(self.nodes_list).index(f"{load['bus']}.{ph}") 
                i_ = get_i(self.syst,'load_' + bus_name,phase_name=n2a[str(ph)],i_type='phasor',dq_name='ri')
                self.I_node[idx] += i_ 
                ph = load['bus_nodes'][1]
                idx = list(self.nodes_list).index(f"{load['bus']}.{ph}") 
                i_ = get_i(self.syst,'load_' + bus_name,phase_name=n2a[str(ph)],i_type='phasor',dq_name='ri')
                self.I_node[idx] += i_ 

        for vsc in self.vscs:
            bus_name = vsc['bus_ac']
            phases = ['a','b','c','n']
            if vsc['type'] == 'ac3ph3wvdcq' or vsc['type'] == 'ac3ph3wpq':
                phases = ['a','b','c']

            for ph in phases:
                idx = list(self.nodes_list).index(f"{vsc['bus_ac']}.{a2n[ph]}") 
                i_ = get_i(self.syst,'vsc_' + bus_name,phase_name=ph,i_type='phasor',dq_name='ri')
                self.I_node[idx] += i_ 
                
            if not vsc['type'] == 'ac3ph3wvdcq'  or vsc['type'] == 'ac3ph3wpq':
                bus_name = vsc['bus_dc']
                for ph in ['a','n']:
                    idx = list(self.nodes_list).index(f"{vsc['bus_dc']}.{a2n[ph]}") 
                    i_ = get_i(self.syst,'vsc_' + bus_name,phase_name=ph,i_type='phasor',dq_name='r')
                    self.I_node[idx] += i_ 
                    
                
        I_lines = self.Y_primitive @ self.A_conect.T @ self.V_node
        self.I_lines = I_lines
        

    def get_v(self):
        '''
		Compute phase-neutral and phase-phase voltages from power flow solution and put values 
		in buses dictionary.		
        '''

        res = {} 
        V_sorted = []
        I_sorted = []
        S_sorted = []
        start_node = 0
        self.V_results = self.V_node
        # self.I_results = self.I_node
        
        V_sorted = self.V_node[self.node_sorter]
        I_sorted = self.I_node[self.node_sorter]   
        
        nodes2string = ['v_an','v_bn','v_cn','v_gn']
        for bus in self.buses:
            N_nodes = bus['N_nodes'] 
#            for node in range(5):
#                bus_node = '{:s}.{:s}'.format(str(bus['bus']),str(node))
#                if bus_node in self.nodes:
#                    V = self.V_results[self.nodes.index(bus_node)][0]
#                    V_sorted += [V]
#                    nodes_in_bus += [node]
#            for node in range(5):
#                bus_node = '{:s}.{:s}'.format(str(bus['bus']),str(node))
#                if bus_node in self.nodes:
#                    I = self.I_results[self.nodes.index(bus_node)][0]
#                    I_sorted += [I]
            if N_nodes==3:   # if 3 phases
                v_ag = V_sorted[start_node+0,0]
                v_bg = V_sorted[start_node+1,0]
                v_cg = V_sorted[start_node+2,0]

                i_a = I_sorted[start_node+0,0]
                i_b = I_sorted[start_node+1,0]
                i_c = I_sorted[start_node+2,0]
                
                s_a = (v_ag)*np.conj(i_a)
                s_b = (v_bg)*np.conj(i_b)
                s_c = (v_cg)*np.conj(i_c)
                
                start_node += 3
                bus.update({'v_an':np.abs(v_ag),
                            'v_bn':np.abs(v_bg),
                            'v_cn':np.abs(v_cg),
                            'v_ng':0.0})
                bus.update({'deg_an':np.angle(v_ag, deg=True),
                            'deg_bn':np.angle(v_bg, deg=True),
                            'deg_cn':np.angle(v_cg, deg=True),
                            'deg_ng':np.angle(0, deg=True)})
                bus.update({'v_ab':np.abs(v_ag-v_bg),
                            'v_bc':np.abs(v_bg-v_cg),
                            'v_ca':np.abs(v_cg-v_ag)})
                bus.update({'p_a':s_a.real,
                            'p_b':s_b.real,
                            'p_c':s_c.real})
                bus.update({'q_a':s_a.imag,
                            'q_b':s_b.imag,
                            'q_c':s_c.imag})
                tup = namedtuple('tup',['v_ag', 'v_bg', 'v_cg'])

                res.update({bus['bus']:tup(v_ag,v_bg,v_cg)})
                
            
            if N_nodes==4:   # if 3 phases + neutral
                v_ag = V_sorted[start_node+0,0]
                v_bg = V_sorted[start_node+1,0]
                v_cg = V_sorted[start_node+2,0]
                v_ng = V_sorted[start_node+3,0]
                i_a = I_sorted[start_node+0,0]
                i_b = I_sorted[start_node+1,0]
                i_c = I_sorted[start_node+2,0]
                i_n = I_sorted[start_node+3,0]  
                
                v_an = v_ag-v_ng
                v_bn = v_bg-v_ng                
                v_cn = v_cg-v_ng
                
                s_a = (v_an)*np.conj(i_a)
                s_b = (v_bn)*np.conj(i_b)
                s_c = (v_cn)*np.conj(i_c)
                bus.update({'v_an':np.abs(v_an),
                            'v_bn':np.abs(v_bn),
                            'v_cn':np.abs(v_cn),
                            'v_ng':np.abs(v_ng)})
                bus.update({'deg_an':np.angle(v_ag-v_ng, deg=True),
                            'deg_bn':np.angle(v_bg-v_ng, deg=True),
                            'deg_cn':np.angle(v_cg-v_ng, deg=True),
                            'deg_ng':np.angle(v_ng, deg=True)})
                bus.update({'v_ab':np.abs(v_ag-v_bg),
                            'v_bc':np.abs(v_bg-v_cg),
                            'v_ca':np.abs(v_cg-v_ag)})
                bus.update({'p_a':s_a.real,
                            'p_b':s_b.real,
                            'p_c':s_c.real})
                bus.update({'q_a':s_a.imag,
                            'q_b':s_b.imag,
                            'q_c':s_c.imag})
    
                start_node += 4

                tup = namedtuple('tup',['v_ag', 'v_bg', 'v_cg', 'v_ng','v_an', 'v_bn', 'v_cn'])

                res.update({bus['bus']:tup(v_ag,v_bg,v_cg,v_ng,v_an,v_bn,v_cn)})

        self.V = np.array(V_sorted).reshape(len(V_sorted),1) 
        self.res = res
        
        return 0 #self.V              
        
        
    def get_i(self):
        '''
		Compute line currents from power flow solution and put values 
		in transformers and lines dictionaries.		
        '''
        I_lines  =self.I_lines
        
        it_single_line = 0
        for trafo in self.transformers:
            
            if 'conductors_j' in trafo: 
                cond_1 = trafo['conductors_j']
            else:
                cond_1 = trafo['conductors_1']
            if 'conductors_k' in trafo: 
                cond_2 = trafo['conductors_k']
            else:
                cond_2 = trafo['conductors_2']  
            
            I_1a = (I_lines[it_single_line,0])
            I_1b = (I_lines[it_single_line+1,0])
            I_1c = (I_lines[it_single_line+2,0])
            I_1n = (I_lines[it_single_line+3,0])
            
            I_2a = (I_lines[it_single_line+cond_1+0,0])
            I_2b = (I_lines[it_single_line+cond_1+1,0])
            I_2c = (I_lines[it_single_line+cond_1+2,0])

            if cond_1>3: I_1n = (I_lines[it_single_line+cond_1+3,0])
            if cond_2>3: I_2n = (I_lines[it_single_line+cond_2+3,0])

            #I_n = (I_lines[it_single_line+3,0])
            if cond_1 <=3:
                I_1n = I_1a+I_1b+I_1c
            if cond_2 <=3:
                I_2n = I_2a+I_2b+I_2c
                
            it_single_line += cond_1 + cond_2
            trafo.update({'i_1a_m':np.abs(I_1a)})
            trafo.update({'i_1b_m':np.abs(I_1b)})
            trafo.update({'i_1c_m':np.abs(I_1c)})
            trafo.update({'i_1n_m':np.abs(I_1n)})
            trafo.update({'i_2a_m':np.abs(I_2a)})
            trafo.update({'i_2b_m':np.abs(I_2b)})
            trafo.update({'i_2c_m':np.abs(I_2c)})
            trafo.update({'i_2n_m':np.abs(I_2n)})
            trafo.update({'deg_1a':np.angle(I_1a, deg=True)})
            trafo.update({'deg_1b':np.angle(I_1b, deg=True)})
            trafo.update({'deg_1c':np.angle(I_1c, deg=True)})
            trafo.update({'deg_1n':np.angle(I_1n, deg=True)})
            trafo.update({'deg_2a':np.angle(I_2a, deg=True)})
            trafo.update({'deg_2b':np.angle(I_2b, deg=True)})
            trafo.update({'deg_2c':np.angle(I_2c, deg=True)})
            trafo.update({'deg_2n':np.angle(I_2n, deg=True)})

                        
        self.I_lines = I_lines
        for line in self.lines:
            if line['type'] == 'z':
                N_conductors = len(line['bus_j_nodes'])
                if N_conductors == 3:
                    I_a = (I_lines[it_single_line,0])
                    I_b = (I_lines[it_single_line+1,0])
                    I_c = (I_lines[it_single_line+2,0])
                    #I_n = (I_lines[it_single_line+3,0])
                    I_n = I_a+I_b+I_c
                    
                    alpha = alpha = np.exp(2.0/3*np.pi*1j)
                    i_z =  1/3*(I_a+I_b+I_c)
                    i_p = 1.0/3.0*(I_a + I_b*alpha + I_c*alpha**2)
                    i_n = 1.0/3.0*(I_a + I_b*alpha**2 + I_c*alpha)                
                    it_single_line += N_conductors
                    line.update({'i_j_a_m':np.abs(I_a)})
                    line.update({'i_j_b_m':np.abs(I_b)})
                    line.update({'i_j_c_m':np.abs(I_c)})
                    line.update({'i_j_n_m':np.abs(I_n)})
                    line.update({'deg_j_a':np.angle(I_a, deg=True)})
                    line.update({'deg_j_b':np.angle(I_b, deg=True)})
                    line.update({'deg_j_c':np.angle(I_c, deg=True)})
                    line.update({'deg_j_n':np.angle(I_n, deg=True)})
                    line.update({'i_k_a_m':np.abs(I_a)})
                    line.update({'i_k_b_m':np.abs(I_b)})
                    line.update({'i_k_c_m':np.abs(I_c)})
                    line.update({'i_k_n_m':np.abs(I_n)})
                    line.update({'deg_k_a':np.angle(I_a, deg=True)})
                    line.update({'deg_k_b':np.angle(I_b, deg=True)})
                    line.update({'deg_k_c':np.angle(I_c, deg=True)})
                    line.update({'deg_k_n':np.angle(I_n, deg=True)})
                    line.update({'i_z':np.abs(i_z)})
                    line.update({'i_p':np.abs(i_p)})
                    line.update({'i_n':np.abs(i_n)})
                if N_conductors == 4:
                    I_a = (I_lines[it_single_line,0])
                    I_b = (I_lines[it_single_line+1,0])
                    I_c = (I_lines[it_single_line+2,0])
                    I_n = (I_lines[it_single_line+3,0])
                    it_single_line += N_conductors
                    line.update({'i_j_a_m':np.abs(I_a)})
                    line.update({'i_j_b_m':np.abs(I_b)})
                    line.update({'i_j_c_m':np.abs(I_c)})
                    line.update({'i_j_n_m':np.abs(I_n)})
                    line.update({'deg_j_a':np.angle(I_a, deg=True)})
                    line.update({'deg_j_b':np.angle(I_b, deg=True)})
                    line.update({'deg_j_c':np.angle(I_c, deg=True)})
                    line.update({'deg_j_n':np.angle(I_n, deg=True)})
                    line.update({'i_k_a_m':np.abs(I_a)})
                    line.update({'i_k_b_m':np.abs(I_b)})
                    line.update({'i_k_c_m':np.abs(I_c)})
                    line.update({'i_k_n_m':np.abs(I_n)})
                    line.update({'deg_k_a':np.angle(I_a, deg=True)})
                    line.update({'deg_k_b':np.angle(I_b, deg=True)})
                    line.update({'deg_k_c':np.angle(I_c, deg=True)})
                    line.update({'deg_k_n':np.angle(I_n, deg=True)})
                    
            if line['type'] == 'pi':
                N_conductors = len(line['bus_j_nodes'])
                if N_conductors == 3:
                    I_j_a = I_lines[it_single_line+0,0]+I_lines[it_single_line+3,0]
                    I_j_b = I_lines[it_single_line+1,0]+I_lines[it_single_line+4,0]
                    I_j_c = I_lines[it_single_line+2,0]+I_lines[it_single_line+5,0]
                    I_k_a = I_lines[it_single_line+0,0]-I_lines[it_single_line+6,0]
                    I_k_b = I_lines[it_single_line+1,0]-I_lines[it_single_line+7,0]
                    I_k_c = I_lines[it_single_line+2,0]-I_lines[it_single_line+8,0]
                    
                    #I_n = (I_lines[it_single_line+3,0])
                    I_j_n = I_j_a+I_j_b+I_j_c
                    I_k_n = I_k_a+I_k_b+I_k_c
                    
                    alpha = alpha = np.exp(2.0/3*np.pi*1j)
                    i_z =  1/3*(I_j_a+I_j_b+I_j_c)
                    i_p = 1.0/3.0*(I_j_a + I_j_b*alpha + I_j_c*alpha**2)
                    i_n = 1.0/3.0*(I_j_a + I_j_b*alpha**2 + I_j_c*alpha)                
                    it_single_line += N_conductors*3
                    line.update({'i_j_a_m':np.abs(I_j_a)})
                    line.update({'i_j_b_m':np.abs(I_j_b)})
                    line.update({'i_j_c_m':np.abs(I_j_c)})
                    line.update({'i_j_n_m':np.abs(I_j_n)})
                    line.update({'deg_j_a':np.angle(I_j_a, deg=True)})
                    line.update({'deg_j_b':np.angle(I_j_b, deg=True)})
                    line.update({'deg_j_c':np.angle(I_j_c, deg=True)})
                    line.update({'deg_j_n':np.angle(I_j_n, deg=True)})
                    line.update({'i_k_a_m':np.abs(I_k_a)})
                    line.update({'i_k_b_m':np.abs(I_k_b)})
                    line.update({'i_k_c_m':np.abs(I_k_c)})
                    line.update({'i_k_n_m':np.abs(I_k_n)})
                    line.update({'deg_k_a':np.angle(I_k_a, deg=True)})
                    line.update({'deg_k_b':np.angle(I_k_b, deg=True)})
                    line.update({'deg_k_c':np.angle(I_k_c, deg=True)})
                    line.update({'deg_k_n':np.angle(I_k_n, deg=True)})
                    line.update({'i_z':np.abs(i_z)})
                    line.update({'i_p':np.abs(i_p)})
                    line.update({'i_n':np.abs(i_n)})
                if N_conductors == 4:
                    I_j_a = I_lines[it_single_line+0,0]+I_lines[it_single_line+3,0]
                    I_j_b = I_lines[it_single_line+1,0]+I_lines[it_single_line+4,0]
                    I_j_c = I_lines[it_single_line+2,0]+I_lines[it_single_line+5,0]
                    I_k_a = I_lines[it_single_line+0,0]-I_lines[it_single_line+6,0]
                    I_k_b = I_lines[it_single_line+1,0]-I_lines[it_single_line+7,0]
                    I_k_c = I_lines[it_single_line+2,0]-I_lines[it_single_line+8,0]
                    I_j_n = I_lines[it_single_line+3,0]
                    I_k_n = I_lines[it_single_line+3,0]
                    
                    #I_n = (I_lines[it_single_line+3,0])
                    I_j_n = I_j_a+I_j_b+I_j_c
                    I_k_n = I_k_a+I_k_b+I_k_c
                    
                    alpha = alpha = np.exp(2.0/3*np.pi*1j)
                    i_z =  1/3*(I_j_a+I_j_b+I_j_c)
                    i_p = 1.0/3.0*(I_j_a + I_j_b*alpha + I_j_c*alpha**2)
                    i_n = 1.0/3.0*(I_j_a + I_j_b*alpha**2 + I_j_c*alpha)                
                    it_single_line += N_conductors*3
                    line.update({'i_j_a_m':np.abs(I_j_a)})
                    line.update({'i_j_b_m':np.abs(I_j_b)})
                    line.update({'i_j_c_m':np.abs(I_j_c)})
                    line.update({'i_j_n_m':np.abs(I_j_n)})
                    line.update({'deg_j_a':np.angle(I_j_a, deg=True)})
                    line.update({'deg_j_b':np.angle(I_j_b, deg=True)})
                    line.update({'deg_j_c':np.angle(I_j_c, deg=True)})
                    line.update({'deg_j_n':np.angle(I_j_n, deg=True)})
                    line.update({'i_k_a_m':np.abs(I_k_a)})
                    line.update({'i_k_b_m':np.abs(I_k_b)})
                    line.update({'i_k_c_m':np.abs(I_k_c)})
                    line.update({'i_k_n_m':np.abs(I_k_n)})
                    line.update({'deg_k_a':np.angle(I_k_a, deg=True)})
                    line.update({'deg_k_b':np.angle(I_k_b, deg=True)})
                    line.update({'deg_k_c':np.angle(I_k_c, deg=True)})
                    line.update({'deg_k_n':np.angle(I_k_n, deg=True)})  
                    

    def bokeh_tools(self):

        
        self.bus_tooltip = '''
            <div>
            bus_id = @bus_id &nbsp &nbsp |  u<sub>avg</sub>= @u_avg_pu pu |  u<sub>unb</sub>= @v_unb %
            <table border="1">
                <tr>
                <td>v<sub>an</sub> =  @v_an  &ang; @deg_an V </td> <td> S<sub>a</sub> = @p_a + j@q_a kVA</td>
                </tr>
                      <tr>
                      <td> </td> <td>v<sub>ab</sub>= @v_ab V</td>
                      </tr>
                <tr>
                <td>v<sub>bn</sub> = @v_bn &ang; @deg_bn V </td><td> S<sub>b</sub> = @p_b + j@q_b kVA</td>
                </tr>
                      <tr>
                      <td> </td><td>v<sub>bc</sub>= @v_bc V</td>
                      </tr>
                <tr>
                <td>v<sub>cn</sub>  = @v_cn &ang; @deg_cn V </td>  <td>S<sub>c</sub> = @p_c + j@q_c kVA </td>
                </tr> 
                    <tr>
                     <td> </td> <td>v<sub>ca</sub>= @v_ca V</td>
                    </tr>
               <tr>
                <td>v<sub>ng</sub>    = @v_ng &ang; @deg_ng V</td>  <td>S<sub>abc</sub> = @p_abc + j@q_abc kVA </td>
              </tr>
            </table>
            </div>
            '''
            
        x = [item['pos_x'] for item in self.buses]
        y = [item['pos_y'] for item in self.buses]
        bus_id = [item['bus'] for item in self.buses]
        v_an = ['{:2.2f}'.format(float(item['v_an'])) for item in self.buses]
        v_bn = ['{:2.2f}'.format(float(item['v_bn'])) for item in self.buses]
        v_cn = ['{:2.2f}'.format(float(item['v_cn'])) for item in self.buses]
        v_ng = ['{:2.2f}'.format(float(item['v_ng'])) for item in self.buses]
        sqrt3=np.sqrt(3)
        
        u_avg_pu = []
        v_unb = []
        for item in  self.buses:
            V_base = float(item['U_kV'])*1000.0/sqrt3
            v_an_float = float(item['v_an'])
            v_bn_float = float(item['v_bn'])
            v_cn_float = float(item['v_cn'])
            v_ng_float = float(item['v_ng'])
            
            v_abc = np.array([v_an_float,v_bn_float,v_cn_float])
            v_avg = np.average(v_abc)
            unb = float(np.max(np.abs(v_abc-v_avg))/v_avg)
            v_avg_pu = float(v_avg/V_base)
            u_avg_pu += ['{:2.3f}'.format(v_avg_pu)]
            v_unb += ['{:2.1f}'.format(unb*100)]
            
        
        v_an_pu = ['{:2.4f}'.format(float(item['v_an'])/float(item['U_kV'])/1000.0*sqrt3) for item in self.buses]
        v_bn_pu = ['{:2.4f}'.format(float(item['v_bn'])/float(item['U_kV'])/1000.0*sqrt3) for item in self.buses]
        v_cn_pu = ['{:2.4f}'.format(float(item['v_cn'])/float(item['U_kV'])/1000.0*sqrt3) for item in self.buses]
        v_ng_pu = ['{:2.4f}'.format(float(item['v_ng'])/float(item['U_kV'])/1000.0*sqrt3) for item in self.buses]
        
        deg_an = ['{:2.2f}'.format(float(item['deg_an'])) for item in self.buses]
        deg_bn = ['{:2.2f}'.format(float(item['deg_bn'])) for item in self.buses]
        deg_cn = ['{:2.2f}'.format(float(item['deg_cn'])) for item in self.buses]
        deg_ng = ['{:2.2f}'.format(float(item['deg_ng'])) for item in self.buses]
        v_ab = [item['v_ab'] for item in self.buses]
        v_bc = [item['v_bc'] for item in self.buses]
        v_ca = [item['v_ca'] for item in self.buses]
        
        p_a = ['{:2.2f}'.format(float(item['p_a']/1000)) for item in self.buses]
        p_b = ['{:2.2f}'.format(float(item['p_b']/1000)) for item in self.buses]
        p_c = ['{:2.2f}'.format(float(item['p_c']/1000)) for item in self.buses]
        q_a = ['{:2.2f}'.format(float(item['q_a']/1000)) for item in self.buses]
        q_b = ['{:2.2f}'.format(float(item['q_b']/1000)) for item in self.buses]
        q_c = ['{:2.2f}'.format(float(item['q_c']/1000)) for item in self.buses]   
        p_abc = ['{:2.2f}'.format(float((item['p_a'] +item['p_b']+item['p_c'])/1000)) for item in self.buses] 
        q_abc = ['{:2.2f}'.format(float((item['q_a'] +item['q_b']+item['q_c'])/1000)) for item in self.buses]
        s_radio = []
        s_color = []
        
        for item in self.buses:
            p_total = item['p_a'] + item['p_b'] + item['p_c']
            q_total = item['q_a'] + item['q_b'] + item['q_c']            
            s_total = np.abs(p_total + 1j*q_total)
            scale = self.s_radio_scale
            s_scaled = abs(np.sqrt(s_total))*scale
            if s_scaled<self.s_radio_min :
                s_scaled = self.s_radio_min 
            if s_scaled>self.s_radio_max:
                s_scaled = self.s_radio_max
            s_radio += [s_scaled]
            if p_total>0.0:
                s_color += ['red']
            if p_total<0.0:
                s_color += ['green']
            if p_total==0.0:
                s_color += ['blue']
                                
        self.bus_data = dict(x=x, y=y, bus_id=bus_id,  u_avg_pu=u_avg_pu,  v_unb=v_unb,
                             v_an=v_an, v_bn=v_bn, v_cn=v_cn, v_ng=v_ng, 
                             v_an_pu=v_an_pu, v_bn_pu=v_bn_pu, v_cn_pu=v_cn_pu, 
                             deg_an=deg_an, deg_bn=deg_bn, deg_cn=deg_cn, 
                             deg_ng=deg_ng,v_ab=v_ab,v_bc=v_bc,v_ca=v_ca,
                             p_a=p_a,p_b=p_b,p_c=p_c,
                             q_a=q_a,q_b=q_b,q_c=q_c,
                             p_abc=p_abc,q_abc=q_abc,
                             s_radio=s_radio, s_color=s_color)
        
        self.line_tooltip = '''
            <div>
            line id = @line_id 
            <table border="1">
                <tr>
                <td>I<sub>a</sub> =  @i_a_m &ang; @deg_a A</td>
                </tr>
                <tr>
                <td>I<sub>b</sub> =  @i_b_m &ang; @deg_b A</td>
                </tr>
                <tr>
                <td>I<sub>c</sub> =  @i_c_m &ang; @deg_c A</td>
                </tr>
                <tr>
                <td>I<sub>n</sub> =  @i_n_m &ang; @deg_n A</td>
                </tr>
            </table>            
            </div>
            '''


        self.line_tooltip = '''
            <div>
            line id = @line_id 
            <table border="5">
                <tr >
                    <td>I<sub>ja</sub> =  @i_j_a_m &ang; @deg_j_a </td>
                    <td>I<sub>ka</sub> =  @i_k_a_m &ang; @deg_k_a </td>
                </tr>
                <tr>
                    <td >I<sub>jb</sub> =  @i_j_b_m &ang; @deg_j_b </td>
                    <td >I<sub>kb</sub> =  @i_k_b_m &ang; @deg_k_b </td>
                </tr>
                <tr>
                    <td >I<sub>jc</sub> =  @i_j_c_m &ang; @deg_j_c </td>
                    <td >I<sub>kc</sub> =  @i_k_c_m &ang; @deg_k_c </td>
                </tr>
                <tr>
                    <td >I<sub>jn</sub> =  @i_j_n_m &ang; @deg_j_n </td>
                    <td >I<sub>kn</sub> =  @i_k_n_m &ang; @deg_k_n </td>
                </tr>
            </table>            
            </div>
            '''

            
        bus_id_to_x = dict(zip(bus_id,x))
        bus_id_to_y = dict(zip(bus_id,y))
        
        x_j = [bus_id_to_x[item['bus_j']] for item in self.lines]
        y_j = [bus_id_to_y[item['bus_j']] for item in self.lines]
        x_k = [bus_id_to_x[item['bus_k']] for item in self.lines]
        y_k = [bus_id_to_y[item['bus_k']] for item in self.lines]
        
        x_s = []
        y_s = []
        for line in self.lines:
            x_s += [[ bus_id_to_x[line['bus_j']] , bus_id_to_x[line['bus_k']]]]
            y_s += [[ bus_id_to_y[line['bus_j']] , bus_id_to_y[line['bus_k']]]]
            
        i_j_a_m = [item['i_j_a_m'] for item in self.lines]
        i_j_b_m = [item['i_j_b_m'] for item in self.lines]
        i_j_c_m = [item['i_j_c_m'] for item in self.lines]
        i_j_n_m = [item['i_j_n_m'] for item in self.lines]
        i_k_a_m = [item['i_k_a_m'] for item in self.lines]
        i_k_b_m = [item['i_k_b_m'] for item in self.lines]
        i_k_c_m = [item['i_k_c_m'] for item in self.lines]
        i_k_n_m = [item['i_k_n_m'] for item in self.lines]
        
        deg_j_a = [item['deg_j_a'] for item in self.lines]
        deg_j_b = [item['deg_j_b'] for item in self.lines]
        deg_j_c = [item['deg_j_c'] for item in self.lines]
        deg_j_n = [item['deg_j_n'] for item in self.lines]
        deg_k_a = [item['deg_k_a'] for item in self.lines]
        deg_k_b = [item['deg_k_b'] for item in self.lines]
        deg_k_c = [item['deg_k_c'] for item in self.lines]
        deg_k_n = [item['deg_k_n'] for item in self.lines] 
        
        line_id = ['{:s}-{:s}'.format(item['bus_j'],item['bus_k']) for item in self.lines]
#        self.line_data = dict(x_j=x_j, x_k=x_k, y_j=y_j, y_k=y_k, line_id=line_id,
#                             i_a_m=i_a_m)
        self.line_data = dict(x_s=x_s, y_s=y_s, line_id=line_id,
                             i_j_a_m=i_j_a_m, i_j_b_m=i_j_b_m, i_j_c_m=i_j_c_m, i_j_n_m=i_j_n_m,
                             i_k_a_m=i_k_a_m, i_k_b_m=i_k_b_m, i_k_c_m=i_k_c_m, i_k_n_m=i_k_n_m,                             
                             deg_j_a=deg_j_a, deg_j_b=deg_j_b, deg_j_c=deg_j_c, deg_j_n=deg_j_n,
                             deg_k_a=deg_k_a, deg_k_b=deg_k_b, deg_k_c=deg_k_c, deg_k_n=deg_k_n)


        
        self.transformer_tooltip = '''
            <div>
            transformer id = @trafo_id  
            <table border="5">
                <tr >
                    <td>I<sub>1a</sub> =  @i_1a_m &ang; @deg_1a </td>
                    <td>I<sub>2a</sub> =  @i_2a_m &ang; @deg_2a </td>
                </tr>
                <tr>
                    <td >I<sub>1b</sub> =  @i_1b_m &ang; @deg_1b </td>
                    <td >I<sub>2b</sub> =  @i_2b_m &ang; @deg_2b </td>
                </tr>
                <tr>
                    <td >I<sub>1c</sub> =  @i_1c_m &ang; @deg_1c </td>
                    <td >I<sub>2c</sub> =  @i_2c_m &ang; @deg_2c </td>
                </tr>
                <tr>
                    <td >I<sub>1n</sub> =  @i_1n_m &ang; @deg_1n </td>
                    <td >I<sub>2n</sub> =  @i_2n_m &ang; @deg_2n </td>
                </tr>
            </table>            
            </div>
            '''
            
        bus_id_to_x = dict(zip(bus_id,x))
        bus_id_to_y = dict(zip(bus_id,y))
        
        x_j = [bus_id_to_x[item['bus_j']] for item in self.transformers]
        y_j = [bus_id_to_y[item['bus_j']] for item in self.transformers]
        x_k = [bus_id_to_x[item['bus_k']] for item in self.transformers]
        y_k = [bus_id_to_y[item['bus_k']] for item in self.transformers]
        
        x_s = []
        y_s = []
        for line in self.transformers:
            x_s += [[ bus_id_to_x[line['bus_j']] , bus_id_to_x[line['bus_k']]]]
            y_s += [[ bus_id_to_y[line['bus_j']] , bus_id_to_y[line['bus_k']]]]
            
        i_1a_m = [item['i_1a_m'] for item in self.transformers]
        i_1b_m = [item['i_1b_m'] for item in self.transformers]
        i_1c_m = [item['i_1c_m'] for item in self.transformers]
        i_1n_m = [item['i_1n_m'] for item in self.transformers]

        i_2a_m = [item['i_2a_m'] for item in self.transformers]
        i_2b_m = [item['i_2b_m'] for item in self.transformers]
        i_2c_m = [item['i_2c_m'] for item in self.transformers]
        i_2n_m = [item['i_2n_m'] for item in self.transformers]
        
        deg_1a = [item['deg_1a'] for item in self.transformers]
        deg_1b = [item['deg_1b'] for item in self.transformers]
        deg_1c = [item['deg_1c'] for item in self.transformers]
        deg_1n = [item['deg_1n'] for item in self.transformers]    
        
        deg_2a = [item['deg_2a'] for item in self.transformers]
        deg_2b = [item['deg_2b'] for item in self.transformers]
        deg_2c = [item['deg_2c'] for item in self.transformers]
        deg_2n = [item['deg_2n'] for item in self.transformers]  
        
        trafo_id = ['{:s}-{:s}'.format(item['bus_j'],item['bus_k']) for item in self.transformers]
#        self.line_data = dict(x_j=x_j, x_k=x_k, y_j=y_j, y_k=y_k, line_id=line_id,
#                             i_a_m=i_a_m)
        self.transformer_data = dict(x_s=x_s, y_s=y_s, trafo_id=trafo_id,
                                     i_1a_m=i_1a_m, i_1b_m=i_1b_m, i_1c_m=i_1c_m, i_1n_m=i_1n_m,
                                     deg_1a=deg_1a, deg_1b=deg_1b, deg_1c=deg_1c, deg_1n=deg_1n,
                                     i_2a_m=i_2a_m, i_2b_m=i_2b_m, i_2c_m=i_2c_m, i_2n_m=i_2n_m,
                                     deg_2a=deg_2a, deg_2b=deg_2b, deg_2c=deg_2c, deg_2n=deg_2n)
        
        return self.bus_data

from bokeh.io import output_notebook, show
from bokeh.plotting import figure
from bokeh.models import ColumnDataSource, HoverTool
from bokeh.io import push_notebook
from bokeh.resources import INLINE
output_notebook(INLINE)



def plot_results(grid):
    #grid.bokeh_tools()
    
    p = figure(width=600, height=400,
               title='Results')
    
    # trafos:
    source = ColumnDataSource(grid.transformer_data)
    trafo = p.multi_line(source=source, xs='x_s', ys='y_s', color="green", alpha=0.5, line_width=5)
    
    # lines:
    source = ColumnDataSource(grid.line_data)
    lin = p.multi_line(source=source, xs='x_s', ys='y_s', color="red", alpha=0.5, line_width=5)
    
    # buses:
    source = ColumnDataSource(grid.bus_data)
    cr = p.circle(source=source, x='x', y='y', size=15, color="navy", alpha=0.5)
    
    p.add_tools(HoverTool(renderers=[trafo], tooltips=grid.transformer_tooltip))
    p.add_tools(HoverTool(renderers=[lin], tooltips=grid.line_tooltip))
    p.add_tools(HoverTool(renderers=[cr], tooltips=grid.bus_tooltip))
    show(p)
    
    return p


def get_flow(grid_obj,bus_j,bus_k,mode='total',model='pydgrid_pydae'):
    if model == 'pydgrid_pydae':
        v_a   = grid_obj.get_values(f'v_{bus_j}_a_r') + 1j* grid_obj.get_values(f'v_{bus_j}_a_i')
        i_l_a = grid_obj.get_values(f'i_l_{bus_j}_{bus_k}_a_r') + 1j* grid_obj.get_values(f'i_l_{bus_j}_{bus_k}_a_i')
        v_b   = grid_obj.get_values(f'v_{bus_j}_b_r') + 1j* grid_obj.get_values(f'v_{bus_j}_b_i')
        i_l_b = grid_obj.get_values(f'i_l_{bus_j}_{bus_k}_b_r') + 1j* grid_obj.get_values(f'i_l_{bus_j}_{bus_k}_b_i')
        v_c   = grid_obj.get_values(f'v_{bus_j}_c_r') + 1j* grid_obj.get_values(f'v_{bus_j}_c_i')
        i_l_c = grid_obj.get_values(f'i_l_{bus_j}_{bus_k}_c_r') + 1j* grid_obj.get_values(f'i_l_{bus_j}_{bus_k}_c_i')
        s_a = v_a*np.conj(i_l_a)
        s_b = v_b*np.conj(i_l_b)
        s_c = v_c*np.conj(i_l_c)

        if mode == 'total':
            s_t = s_a + s_b + s_c
            return s_t
        if mode == 'abc':
            return s_a,s_b,s_c
        
def set_voltage(grid_obj,bus_name,voltage,phase):
    '''
    Set new power to a grid feeder.

    Parameters
    ----------
    grid_obj : object of pydgrid.grid class
    bus_name : string
        name of the grid feeder bus.
    voltage : real escalar
        phase-phase RMS voltage magnitude
    phase : real escalar.
        phase angle in degree.

    Returns
    -------
    None.

    '''

 
    v_a = voltage/np.sqrt(3)*np.exp(1j*np.deg2rad(phase))
    v_b = voltage/np.sqrt(3)*np.exp(1j*np.deg2rad(phase-240))
    v_c = voltage/np.sqrt(3)*np.exp(1j*np.deg2rad(phase-120))
    grid_obj.set_value(f'v_{bus_name}_a_r',v_a.real)
    grid_obj.set_value(f'v_{bus_name}_a_i',v_a.imag)
    grid_obj.set_value(f'v_{bus_name}_b_r',v_b.real)
    grid_obj.set_value(f'v_{bus_name}_b_i',v_b.imag)
    grid_obj.set_value(f'v_{bus_name}_c_r',v_c.real)
    grid_obj.set_value(f'v_{bus_name}_c_i',v_c.imag)
 
def set_voltages(grid_obj,bus_name,voltages,phases):
    '''
    Set new power to a grid feeder.

    Parameters
    ----------
    grid_obj : object of pydgrid.grid class
    bus_name : string
        name of the grid feeder bus.
    voltage : real escalar
        phase-phase RMS voltage magnitude
    phase : real escalar.
        phase angle in degree.

    Returns
    -------
    None.

    '''

    if isinstance(phases, list):
        v_a = voltages[0]*np.exp(1j*np.deg2rad(phases[0]))
        v_b = voltages[1]*np.exp(1j*np.deg2rad(phases[1]))
        v_c = voltages[2]*np.exp(1j*np.deg2rad(phases[2]))
    else:
        v_a = voltages[0]*np.exp(1j*np.deg2rad(phases))
        v_b = voltages[1]*np.exp(1j*np.deg2rad(phases-240))
        v_c = voltages[2]*np.exp(1j*np.deg2rad(phases-120))    
    
    grid_obj.set_value(f'v_{bus_name}_a_r',v_a.real)
    grid_obj.set_value(f'v_{bus_name}_a_i',v_a.imag)
    grid_obj.set_value(f'v_{bus_name}_b_r',v_b.real)
    grid_obj.set_value(f'v_{bus_name}_b_i',v_b.imag)
    grid_obj.set_value(f'v_{bus_name}_c_r',v_c.real)
    grid_obj.set_value(f'v_{bus_name}_c_i',v_c.imag)
    
    
def phasor2inst(grid_obj,bus_name,magnitude='v',to_bus='',phases=['a','b','c'],Freq = 50,Dt=1e-4):
    
    omega = 2*np.pi*Freq
    out = []
    
    if magnitude == 'v':
        for ph in phases:
            Times = np.arange(0.0,grid_obj.T[-1,0],Dt)
            R = grid_obj.get_values(f'{magnitude}_{bus_name}_{ph}_r') 
            I = grid_obj.get_values(f'{magnitude}_{bus_name}_{ph}_i')
            R_ = np.interp(Times,grid_obj.T[:,0],R)
            I_ = np.interp(Times,grid_obj.T[:,0],I)
            R_I = R_ + 1j*I_
            cplx = np.sqrt(2)*np.exp(1j*omega*Times)*R_I
            out += [cplx.real]

    if magnitude == 'iline':
        for ph in phases:
            Times = np.arange(0.0,grid_obj.T[-1,0],Dt)
            R = grid_obj.get_values(f'i_l_{bus_name}_{to_bus}_{ph}_r') 
            I = grid_obj.get_values(f'i_l_{bus_name}_{to_bus}_{ph}_i')
            R_ = np.interp(Times,grid_obj.T[:,0],R)
            I_ = np.interp(Times,grid_obj.T[:,0],I)
            R_I = R_ + 1j*I_
            cplx = np.sqrt(2)*np.exp(1j*omega*Times)*R_I
            out += [cplx.real]

    return Times,out


def get_voltage(grid_obj,bus_name,output='v_an_m'):
    '''
    Get voltage module of a bus.

    Parameters
    ----------
    grid_obj : object of pydae class
    bus_name : string
        name of the bus.
    output : string
         v_an: a phase to neutral voltage phasor (V).
         v_an_m: a phase to neutral RMS voltage (V).
         v_abcn_m: a,b and c phases to neutral voltage phasors (V).
        
    Returns
    -------
    phase-ground voltage module (V).

    '''
    
    if output in ['v_an','v_bn','v_cn']:
        v_sub = f'v_{bus_name}_{output[-2]}'
        v = grid_obj.get_value(f'{v_sub}_r') + 1j* grid_obj.get_value(f'{v_sub}_i')
        return v
    
    if output in ['v_an_m','v_bn_m','v_cn_m']:
        v_sub = f'v_{bus_name}_{output[-2]}'[:-2]
        v = grid_obj.get_value(f'{v_sub}_r') + 1j* grid_obj.get_value(f'{v_sub}_i')
        return np.abs(v)
    
    if output in ['v_abcn']:
        v_list = []
        for ph in ['a','b','c']:
            v_sub = f'v_{bus_name}_{ph}'
            v = grid_obj.get_value(f'{v_sub}_r') + 1j* grid_obj.get_value(f'{v_sub}_i')
            v_list += [v]
        return np.array(v_list).reshape((3,1))
    


@numba.njit(cache=True)
def abc2pq(times,v_a,v_b,v_c,i_a,i_b,i_c,omega=2*np.pi*50,theta_0=0.0):
    N_t = len(times)
    Dt = times[1]-times[0] 
    p = np.zeros((N_t,1))
    q = np.zeros((N_t,1))
    for it in range(len(times)):

        theta = Dt*it*omega + theta_0
        v_abc = np.array([[v_a[it]],[v_b[it]],[v_c[it]]])
        T_p = 2.0/3.0*np.array([[ np.cos(theta), np.cos(theta-2.0/3.0*np.pi), np.cos(theta+2.0/3.0*np.pi)],
                                [-np.sin(theta),-np.sin(theta-2.0/3.0*np.pi),-np.sin(theta+2.0/3.0*np.pi)]])

        dq=T_p@v_abc;
        
        v_d = dq[0]
        v_q = dq[1]

        theta = Dt*it*omega + theta_0
        i_abc = np.array([[i_a[it]],[i_b[it]],[i_c[it]]])
        T_p = 2.0/3.0*np.array([[ np.cos(theta), np.cos(theta-2.0/3.0*np.pi), np.cos(theta+2.0/3.0*np.pi)],
                                [-np.sin(theta),-np.sin(theta-2.0/3.0*np.pi),-np.sin(theta+2.0/3.0*np.pi)]])

        i_dq=T_p@i_abc;
        
        i_d = i_dq[0]
        i_q = i_dq[1]
    
        p[it] = 3/2*(v_d*i_d + v_q*i_q)
        q[it] = 3/2*(v_d*i_q - v_q*i_d)
    return p,q


@numba.njit(cache=True)
def abc2dq(times,v_a,v_b,v_c,i_a,i_b,i_c,omega=2*np.pi*50,theta_0=0.0,K_p=0.1,K_i=20.0,T_f=20.0e-3):

    N_t = len(times)
    Dt = times[1]-times[0] 

    v_d = np.zeros((N_t,1))
    v_q = np.zeros((N_t,1))
    i_d = np.zeros((N_t,1))
    i_q = np.zeros((N_t,1))
    p = np.zeros((N_t,1))
    q = np.zeros((N_t,1))
    
    theta = 0.0
    xi = 0.0
    theta_pll = np.zeros((N_t,1)) 
    omega_pll = np.zeros((N_t,1)) 
    dq = np.zeros((2,1)) 
    idx = np.argmax(times>0.08)

    theta_pll[0,0] = theta_0
    #omega_pll = np.zeros((N_t,1)) 
    
    
    for it in range(len(times)-1):

        theta = theta_pll[it,0]
        v_abc = np.array([[v_a[it]],[v_b[it]],[v_c[it]]])
        i_abc = np.array([[i_a[it]],[i_b[it]],[i_c[it]]])
        T_p = 2.0/3.0*np.array([[ np.cos(theta), np.cos(theta-2.0/3.0*np.pi), np.cos(theta+2.0/3.0*np.pi)],
                                [-np.sin(theta),-np.sin(theta-2.0/3.0*np.pi),-np.sin(theta+2.0/3.0*np.pi)]])
    

        v_dq = T_p@v_abc;
        i_dq = T_p@i_abc;
        
        v_d[it+1,0] = v_d[it,0] + Dt/T_f*(v_dq[0,0] - v_d[it,0])
        v_q[it+1,0] = v_q[it,0] + Dt/T_f*(v_dq[1,0] - v_q[it,0])
        i_d[it+1,0] = i_d[it,0] + Dt/T_f*(i_dq[0,0] - i_d[it,0])
        i_q[it+1,0] = i_q[it,0] + Dt/T_f*(i_dq[1,0] - i_q[it,0])
        
        p[it] = 3/2*(v_d[it+1,0] *i_d[it+1,0]  + v_q[it+1,0] *i_q[it+1,0] )
        q[it] = 3/2*(v_q[it+1,0] *i_d[it+1,0]  - v_d[it+1,0] *i_q[it+1,0] )
        
        xi += Dt*v_dq[0,0]
        omega_pll[it,0] = K_p * v_dq[0,0] + K_i * xi + omega
        
        theta_pll[it+1,0] += theta_pll[it,0] + Dt*(omega_pll[it,0])

    omega_pll[it+1,0] = K_p * v_dq[0,0] + K_i * xi + omega
    return theta_pll,omega_pll,v_d,v_q,i_d,i_q,p,q


def change_line(system,bus_j,bus_k, *args,**kwagrs):
    line = kwagrs
    S_base = system.get_value('S_base')
    
    line_name = f"{bus_j}_{bus_k}"
    if 'X_pu' in line:
        if 'S_mva' in line: S_line = 1e6*line['S_mva']
        R = line['R_pu']*S_base/S_line  # in pu of the system base
        X = line['X_pu']*S_base/S_line  # in pu of the system base
    if 'X' in line:
        U_base = system.get_value(f'U_{bus_j}_n') 
        Z_base = U_base**2/S_base
        R = line['R']/Z_base  # in pu of the system base
        X = line['X']/Z_base  # in pu of the system base
    if 'X_km' in line:
        U_base = system.get_value(f'U_{bus_j}_n')
        Z_base = U_base**2/S_base
        R = line['R_km']*line['km']/Z_base  # in pu of the system base
        X = line['X_km']*line['km']/Z_base  # in pu of the system base
    if 'Bs_km' in line:
        U_base = system.get_value(f'U_{bus_j}_n')
        Z_base = U_base**2/S_base
        print('U_base',U_base,'Z_base',Z_base)
        Y_base = 1.0/Z_base
        Bs = line['Bs_km']*line['km']/Y_base  # in pu of the system base
        bs = Bs
        system.set_value(f'bs_{line_name}',bs)
        print(bs)
    G =  R/(R**2+X**2)
    B = -X/(R**2+X**2)
    system.set_value(f"g_{line_name}",G)
    system.set_value(f"b_{line_name}",B)
    
    
def get_line_i(system,bus_from,bus_to,U_kV=66e3):
    
    
    if f"b_{bus_from}_{bus_to}" in system.params_list: 
        bus_j = bus_from
        bus_k = bus_to
        current_direction = 1.0
    elif f"b_{bus_to}_{bus_from}" in system.params_list: 
        bus_j = bus_to
        bus_k = bus_from
        current_direction = -1.0
    else: 
        print(f'No line from {bus_from} to {bus_to}')
        return
    
    line_name = f"{bus_j}_{bus_k}"
    V_j_m = system.get_value(f"V_{bus_j}")
    theta_j = system.get_value(f"theta_{bus_j}")
    V_k_m = system.get_value(f"V_{bus_k}") 
    theta_k = system.get_value(f"theta_{bus_k}")
    
    V_j = V_j_m*np.exp(1j*theta_j)
    V_k = V_k_m*np.exp(1j*theta_k)
    
    Y_jk = system.get_value(f"g_{line_name}") + 1j*system.get_value(f"b_{line_name}")
    S_base = system.get_value('S_base')
    U_base = system.get_value(f"U_{bus_j}_n")
    I_jk_pu = current_direction*Y_jk*(V_j - V_k)
    I_base = S_base/(np.sqrt(3)*U_base)
    I_jk = I_jk_pu*I_base
    
    return I_jk

def get_line_s(system,bus_from,bus_to,U_kV=66e3):
    
    
    if f"b_{bus_from}_{bus_to}" in system.params_list: 
        bus_j = bus_from
        bus_k = bus_to
        current_direction = 1.0
    elif f"b_{bus_to}_{bus_from}" in system.params_list: 
        bus_j = bus_to
        bus_k = bus_from
        current_direction = -1.0
    else: 
        print(f'No line from {bus_from} to {bus_to}')
        return
    
    line_name = f"{bus_j}_{bus_k}"
    V_j_m = system.get_value(f"V_{bus_j}")
    theta_j = system.get_value(f"theta_{bus_j}")
    V_k_m = system.get_value(f"V_{bus_k}") 
    theta_k = system.get_value(f"theta_{bus_k}")
    
    V_j = V_j_m*np.exp(1j*theta_j)
    V_k = V_k_m*np.exp(1j*theta_k)
    
    Y_jk = system.get_value(f"g_{line_name}") + 1j*system.get_value(f"b_{line_name}")
    S_base = system.get_value('S_base')
    U_base = system.get_value(f"U_{bus_j}_n")
    I_jk_pu = current_direction*Y_jk*(V_j - V_k)
    I_base = S_base/(np.sqrt(3)*U_base)
    I_jk = I_jk_pu*I_base
    S_jk_pu = V_j*np.conj(I_jk_pu)
    S_jk = S_base*S_jk_pu
    
    return S_jk


def set_powers(grid_obj,bus_name,s_cplx,mode='urisi_3ph'):
    '''
    Function for simplifying the power setting.

    Parameters
    ----------
    grid_obj : pydae object
        pydae object with grid of type pydgrid.
    bus_name : string
        name of the bus to xhange the power.
    s_cplx : TYPE
        complex power (negative for load, positive for generated).

    Returns
    -------
    None.

    '''
    
    if mode == 'urisi_3ph':
        p = s_cplx.real
        q = s_cplx.imag
        for ph in ['a','b','c']:
            grid_obj.set_value(f'p_{bus_name}_{ph}',p/3)
            grid_obj.set_value(f'q_{bus_name}_{ph}',q/3)
        
    if mode == 'urisi_abc':
        p = s_cplx.real
        q = s_cplx.imag
        for ph in ['a','b','c']:
            grid_obj.set_value(f'p_{bus_name}_{ph}',p/3)
            grid_obj.set_value(f'q_{bus_name}_{ph}',q/3)  
            
def update_loads(grid,data_input):
    
    if type(data_input) == dict:
        data = data_input
        
    if type(data_input) == str:
        data_input = open(data_input).read().replace("'",'"')
        data = json.loads(data_input)
        
    if 'loads' in data:
        for load in data['loads']:
            
            if load['type'] == "1P+N":
                bus = load['bus']
                kVA = load['kVA']
                pf = load['pf']
                p = kVA*1000*pf
                q = np.sqrt((kVA*1000)**2 - p**2)*np.sign(pf)

                grid.set_value(f'p_load_{bus}_1',p)
                grid.set_value(f'q_load_{bus}_1',q)

def report_v(grid,data_input,show=True,model='urisi'):
    '''
    

    Parameters
    ----------
    grid : pydae object
        Pydae object modelling a grid.
    data_input : if dict, a dictionary with the grid parameters
                 if string, the path to the .json file containing grid parameters

    show : string, optional
        If report is print or not.
        
    model : string, optional
        Type of implemented model. The default is 'urisi'.

    Returns
    -------
    dict with the results.

    '''
    
    
    if type(data_input) == dict:
        data = data_input
        
    if type(data_input) == str:
        data_input = open(data_input).read().replace("'",'"')
        data = json.loads(data_input)
        
    buses_dict = {}
        
    for bus in data['buses']:
        if f"v_{bus['bus']}_n_r" in grid.y_ini_list:
            v_n_r,v_n_i = grid.get_mvalue([f"v_{bus['bus']}_n_r",f"v_{bus['bus']}_n_i"])
            v_n_g = v_n_r + 1j*v_n_i
        else:
            v_n_g = 0.0
        #for ph in ['a','b','c']:
        
        ph = 'a'
        v_a_r,v_a_i =  grid.get_mvalue([f"v_{bus['bus']}_{ph}_r",f"v_{bus['bus']}_{ph}_i"])
        v_a_g = v_a_r + 1j*v_a_i
        v_a_n = v_a_g - v_n_g
        v_a_m = np.abs(v_a_n)
        v_a_a = np.rad2deg(np.angle(v_a_n))

        ph = 'b'
        v_b_r,v_b_i =  grid.get_mvalue([f"v_{bus['bus']}_{ph}_r",f"v_{bus['bus']}_{ph}_i"])
        v_b_g = v_b_r + 1j*v_b_i
        v_b_n = v_b_g - v_n_g
        v_b_m = np.abs(v_b_n)
        v_b_a = np.rad2deg(np.angle(v_b_n))

        ph = 'c'
        v_c_r,v_c_i =  grid.get_mvalue([f"v_{bus['bus']}_{ph}_r",f"v_{bus['bus']}_{ph}_i"])
        v_c_g = v_c_r + 1j*v_c_i
        v_c_n = v_c_g - v_n_g
        v_c_m = np.abs(v_c_n)
        v_c_a = np.rad2deg(np.angle(v_c_n))
 
        alpha = alpha = np.exp(2.0/3*np.pi*1j)
        v_0 =  1/3*(v_a_g+v_b_g+v_c_g)
        v_1 = 1.0/3.0*(v_a_g + v_b_g*alpha + v_c_g*alpha**2)
        v_2 = 1.0/3.0*(v_a_g + v_b_g*alpha**2 + v_c_g*alpha)
        
        # compute unbalanced as in Kersting 1ers edition pg. 266
        v_m_array = [v_a_m,v_b_m,v_c_m]
        v_m_min = np.min(v_m_array)
        v_m_max = np.max(v_m_array)
        max_dev = v_m_max - v_m_min
        v_avg = np.sum(v_m_array)/3
        unbalance = max_dev/v_avg
        
        bus[f"v_{bus['bus']}_{'a'}n"] = v_a_m
        
        if show:
            print(f"v_{bus['bus']}_{'a'}n: {v_a_m:7.1f}| {v_a_a:6.1f}º V,    v_{bus['bus']}_{'a'}g: {np.abs(v_a_g):7.1f}| {np.angle(v_a_g,deg=True):6.1f}º V,    v_1 = {np.abs(v_1):7.1f} V, unb = {unbalance*100:3.2f}%")
            print(f"v_{bus['bus']}_{'b'}n: {v_b_m:7.1f}| {v_b_a:6.1f}º V,    V_{bus['bus']}_{'b'}g: {np.abs(v_b_g):7.1f}| {np.angle(v_b_g,deg=True):6.1f}º V,    v_2 = {np.abs(v_2):7.1f} V")
            print(f"v_{bus['bus']}_{'c'}n: {v_c_m:7.1f}| {v_c_a:6.1f}º V,    V_{bus['bus']}_{'c'}g: {np.abs(v_c_g):7.1f}| {np.angle(v_c_g,deg=True):6.1f}º V,    v_0 = {np.abs(v_0):7.1f} V")                   
            
            print(f"  V_{bus['bus']}_ng: {np.abs(v_n_g):8.1f}| {np.angle(v_n_g,deg=True):8.1f}º V")
        
        buses_dict[bus['bus']] = {f"v_{bus['bus']}_{'a'}n":v_b_m,f"v_{bus['bus']}_{'b'}n":v_b_m,f"v_{bus['bus']}_{'c'}n":v_c_m}
        buses_dict[bus['bus']].update({f'v_unb':unbalance,'v_ng':np.abs(v_n_g)})

    return buses_dict        
        
def report_trafos(grid,data_input,model='urisi'):
    '''
    

    Parameters
    ----------
    grid : pydae object
        Pydae object modelling a grid.
    data_input : if dict, a dictionary with the grid parameters
                 if string, the path to the .json file containing grid parameters

    model : string, optional
        Type of implemented model. The default is 'urisi'.

    Returns
    -------
    None.

    '''
    
    
    if type(data_input) == dict:
        data = data_input
        
    if type(data_input) == str:
        data_input = open(data_input).read().replace("'",'"')
        data = json.loads(data_input)
        
    for trafo in data['transformers']:
        
        bus_j_name = trafo['bus_j']
        bus_k_name = trafo['bus_k']
        
        for ph in ['a','b','c']:
            i_r,i_i = grid.get_mvalue([f"i_t_{bus_j_name}_{bus_k_name}_1_{ph}_r",
                                       f"i_t_{bus_j_name}_{bus_k_name}_1_{ph}_i"])
            i_1 = i_r + 1j*i_i
            i_m = np.abs(i_1)
            phi = np.arctan2(i_i,i_r)
            print(f"I_1_{ph}: {i_m:8.1f} |{np.rad2deg(phi):6.1f}º A")
            
        for ph in ['a','b','c']:
            i_r,i_i = grid.get_mvalue([f"i_t_{bus_j_name}_{bus_k_name}_2_{ph}_r",
                                       f"i_t_{bus_j_name}_{bus_k_name}_2_{ph}_i"])
            i_1 = i_r + 1j*i_i
            i_m = np.abs(i_1)
            phi = np.arctan2(i_i,i_r)
            print(f"I_2_{ph}: {i_m:8.1f} |{np.rad2deg(phi):6.1f}º A")
            
        if f"i_t_{bus_j_name}_{bus_k_name}_2_n_r" in grid.outputs_list:
            i_r,i_i = grid.get_mvalue([f"i_t_{bus_j_name}_{bus_k_name}_2_n_r",
                                       f"i_t_{bus_j_name}_{bus_k_name}_2_n_i"])
            i_1 = i_r + 1j*i_i
            i_m = np.abs(i_1)
            phi = np.arctan2(i_i,i_r)            
            print(f"I_2_n: {i_m:8.1f} |{np.rad2deg(phi):6.1f}º A")


def load_shape(grid,data_input,model='urisi'):
    '''
    

    Parameters
    ----------
    grid : pydae object
        Pydae object modelling a grid.
    data_input : if dict, a dictionary with the grid parameters
                 if string, the path to the .json file containing grid parameters

    model : string, optional
        Type of implemented model. The default is 'urisi'.

    Returns
    -------
    None.

    '''
    
    
    if type(data_input) == dict:
        data = data_input
        
    if type(data_input) == str:
        data_input = open(data_input).read().replace("'",'"')
        data = json.loads(data_input)
        
def read_shapes(data_input):
    '''
    

    Parameters
    ----------
    data_input : if string, the path to the .json file containing grid parameters

    Returns
    -------
    dict with the shapes.

    '''
    
    data_input = open(data_input).read().replace("'",'"')
    data = json.loads(data_input)
    shapes = {}
    for item in data['shapes']:
        values = np.array(data['shapes'][item])
        shapes.update({item:{'t':values[:,0],'val':values[:,1]}})
        
    
    
    return shapes


@numba.njit(cache=True) 
def daesolver(struct): 
    sin = np.sin
    cos = np.cos
    sqrt = np.sqrt
    i = 0 
    
    Dt = struct[i].Dt 
    N_steps = struct[i].N_steps 
    N_store = struct[i].N_store 
    N_x = struct[i].N_x
    N_y = struct[i].N_y
    N_outs = 1 
    decimation = struct[i].decimation 
    eye = np.eye(N_x)
    # initialization 
    #t = struct[i].t 
    t = 0.0
    run(0.0,struct, 1) 
    it_store = 0 
    struct[i]['T'][0] = t 
    struct[i].X[0,:] = struct[i].x[:,0]  
    struct[i].Y[0,:] = struct[i].y[:,0]  


    solver = struct[i].solvern 
    for it in range(N_steps-1): 
        t += Dt 
 
        perturbations(t,struct) 
        
        if solver == 1: 
            # forward euler solver  
            run(t,struct, 2)  
            struct[i].x[:] += Dt*struct[i].f  
 
        if solver == 2: 
            
            # bacward euler solver
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct, 10)  
                phi =  x_0 + Dt*struct[i].f - struct[i].x 
                Dx = np.linalg.solve(-(Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                if np.max(np.abs(Dx)) < struct[i].itol: break 
            print(struct[i].f)
 
        if solver == 3: 
            # trapezoidal solver
            run(t,struct, 2) 
            f_0 = np.copy(struct[i].f[:]) 
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run(t,struct, 10)  
                phi =  x_0 + 0.5*Dt*(f_0 + struct[i].f) - struct[i].x 
                Dx = np.linalg.solve(-(0.5*Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                run(t,struct, 2) 
                if np.max(np.abs(Dx)) < struct[i].itol: break 

        if solver == 4: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

                if np.max(np.abs(Dxy_i[:,0]))<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
            
        if solver == 5: # Teapezoidal DAE as in Milano's book

            run(t,struct, 2) 
            run(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run(t,struct, 2) 
                run(t,struct, 3) 
                run(t,struct,10) 
                run(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                # print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

        # [f_i,g_i,F_x_i,F_y_i,G_x_i,G_y_i] =  smib_transient(x_i,y_i,u);
        
        # A_c_i = [[eye(N_x)-0.5*Dt*F_x_i, -0.5*Dt*F_y_i],
        #          [                G_x_i,         G_y_i]];
             
        # f_n_i = x_i - x - 0.5*Dt*(f_i+f);
        
        # Dxy_i = -A_c_i\[f_n_i.',g_i.'].';
        
        # x_i = x_i + Dxy_i(1:N_x);
        # y_i = y_i + Dxy_i(N_x+1:N_x+N_y);
                
                xy = np.vstack((x_i,y_i))
                max_relative = 0.0
                for it_var in range(N_x+N_y):
                    abs_value = np.abs(xy[it_var,0])
                    if abs_value < 0.001:
                        abs_value = 0.001
                                             
                    relative_error = np.abs(Dxy_i[it_var,0])/abs_value
                    
                    if relative_error > max_relative: max_relative = relative_error
                    
                if max_relative<struct[i].itol:
                    
                    break
                
                # if iter>struct[i].imax-2:
                    
                #     print('Convergence problem')

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
                
        # channels 
        if it >= it_store*decimation and it_store < N_store-1: 
            struct[i]['T'][it_store+1] = t 
            struct[i].X[it_store+1,:] = struct[i].x[:,0] 
            struct[i].Y[it_store+1,:] = struct[i].y[:,0]
            it_store += 1 
    struct[i].t = t
    return struct[i]['T'][:], struct[i].X[:]


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 14 11:53:28 2018

@author: jmmauricio

bug: if there is not jacobian add pass to the if mode == 12:
"""
import numpy as np
import sympy as sym
from collections import deque 
import pkgutil
import re
from sympy.matrices.sparsetools import _doktocsr
from sympy import SparseMatrix
import time
def sym_gen_str():

    str = '''\
import sympy as sym

if 'params' in sys_vars: params = sys_vars['params']
if 'params_dict' in sys_vars: params = sys_vars['params_dict']
u_list = sys_vars['u_list']
x_list = sys_vars['x_list']
y_list = sys_vars['y_list']
for item in params:
    exec(f"{item} = sym.Symbol('{item}')", globals())
for item in u_list:
    exec(f"{item} = sym.Symbol('{item}')", globals())
for item in x_list:
    exec(f"{item} = sym.Symbol('{item}')", globals())
for item in y_list:
    exec(f"{item} = sym.Symbol('{item}')", globals())  
sym_func_list = ['sin','cos']
for item in sym_func_list:
    exec(f"{item} = sym.{item}", globals()) 
    '''
    return str

def sym_gen(sys_vars):
    """Generates global sympy symbolic variables from a dictionary.
    
    Parameters
    ----------
    sys_vars : A ``dict`` with the following keys:
    
    params: ``dict``: Dictionary of the parameters and their values.
    u_list: ``list(str)``: List of input variables as strings for the running problem. 
    x_list: ``list(str)``: List of dynamical states as strings for the running problem.
    y_list: ``list(str)``: List of algebraic states as strings for the running problem.   

    R = 0.1
    L = 0.01
    T_pi = L/R
    K_p = L/0.01

    params= {'R' : 0.1, 'L':0.01, 'V_max':10.0, 'V_min':0.0, 'K_p':1.0, 'T_pi':0.01, 'K_aw':1.0}
    u_ini_dict = { 'i_ref':1.0}  # for the initialization problem
    u_run_dict = { 'i_ref':1.0}  # for the running problem (here initialization and running problem are the same)

    x_list = ['i','xi']
    y_ini_list = ['y','v']
    y_run_list = ['y','v']

    sys_vars = {'params':params,
                'u_list':u_run_dict,
                'x_list':x_list,
                'y_run_list':y_run_list}   
    """    

    params = sys_vars['params_dict']
    u_list = sys_vars['u_list']
    x_list = sys_vars['x_list']
    y_list = sys_vars['y_list']

    for item in params:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())
    for item in u_list:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())
    for item in x_list:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())
    for item in y_list:
        exec(f"{item} = sym.Symbol('{item}', real=True)", globals())  
    sym_func_list = ['sin','cos']
    for item in sym_func_list:
        exec(f"{item} = sym.{item}", globals()) 


    return sys_vars

    
def check_system(sys):
    
    if len(sys['f_list']) == 0:
        print('system without dynamic equations, adding dummy dynamic equation')
        x_dummy,u_dummy = sym.symbols('x_dummy,u_dummy')
        sys['x_list'] = ['x_dummy']
        sys['f_list'] = [u_dummy-x_dummy]
        sys['u_ini_dict'].update({'u_dummy':1.0})
        sys['u_run_dict'].update({'u_dummy':1.0})

    if len(sys['g_list']) == 0:
        print('system without algebraic equations, adding dummy algebraic equation')
        y_dummy,u_dummy = sym.symbols('y_dummy,u_dummy')
        
        sys['g_list'] = [u_dummy-y_dummy]
        sys['y_ini_list'] = ['y_dummy']
        sys['y_run_list'] = ['y_dummy']
        sys['u_ini_dict'].update({'u_dummy':1.0})
        sys['u_run_dict'].update({'u_dummy':1.0})
        
    y_ini_set = set(sys['y_ini_list'])
    contains_duplicates = len(sys['y_ini_list']) != len(y_ini_set)
    if contains_duplicates:
        print('error: y_ini contains duplicates')

    y_set = set(sys['y_run_list'])
    contains_duplicates = len(sys['y_run_list']) != len(y_set)
    if contains_duplicates:
        print('error: y_run contains duplicates')
       
def system(sys, verbose=False):
    '''
    

    Parameters
    ----------
    sys : dict
           {'name':sys_name,
            'params_dict':params_dict,
            'f_list':f_list,
            'g_list':g_list,
            'x_list':x_list,
            'y_ini_list':y_ini_list,
            'y_run_list':y_run_list,
            'u_run_dict':u_run_dict,
            'u_ini_dict':u_ini_dict,
            'h_dict':h_dict}

    Returns
    -------
    sys : TYPE
        DESCRIPTION.

    '''
    
    t_0 = time.time()
    if verbose: print(f'check_system (time: {time.time()-t_0})')
    check_system(sys)
    
    f = sym.Matrix(sys['f_list']).T
    g = sym.Matrix(sys['g_list']).T
    x = sym.Matrix(sys['x_list']).T
    y_ini = sym.Matrix(sys['y_ini_list']).T
    y_run = sym.Matrix(sys['y_run_list']).T
    u_ini = sym.Matrix(list(sys['u_ini_dict'].keys()), real=True)
    
    u_run_list = [sym.Symbol(item,real=True) for item in list(sys['u_run_dict'].keys())]
    u_run = sym.Matrix(u_run_list).T 
    h =  sym.Matrix(list(sys['h_dict'].values())).T     

    if verbose: print(f'computing jacobians Fx_run,Fy_run  (time: {time.time()-t_0})')
    Fx_run = f.jacobian(x)
    Fy_run = f.jacobian(y_run)
    if verbose: print(f'computing jacobians Gx_run,Gy_run  (time: {time.time()-t_0})')
    Gx_run = g.jacobian(x)
    Gy_run = g.jacobian(y_run)

    if verbose: print(f'computing jacobians Fu_run,Gu_run  (time: {time.time()-t_0})')
    Fu_run = f.jacobian(u_run)
    Gu_run = g.jacobian(u_run)

    if verbose: print(f'computing jacobians Fx_ini,Fy_ini  (time: {time.time()-t_0}')   
    Fx_ini = f.jacobian(x)
    Fy_ini = f.jacobian(y_ini)
    if verbose: print(f'computing jacobians Gx_ini,Gy_ini  (time: {time.time()-t_0}')
    Gx_ini = g.jacobian(x)
    Gy_ini = g.jacobian(y_ini)

    if verbose: print(f'computing jacobians Hx_run,Hy_run,Hu_run  (time: {time.time()-t_0})')   
    Hx_run = h.jacobian(x)
    Hy_run = h.jacobian(y_run)   
    Hu_run = h.jacobian(u_run)

    sys['f']= f
    sys['g']= g 
    sys['x']= x
    sys['y_ini']= y_ini  
    sys['y_run']= y_run      
    sys['u_ini']= u_ini 
    sys['u_run']= u_run
    sys['h']= h 
    
    sys['Fx_run'] = Fx_run
    sys['Fy_run'] = Fy_run
    sys['Gx_run'] = Gx_run
    sys['Gy_run'] = Gy_run

    sys['Fx_ini'] = Fx_ini
    sys['Fy_ini'] = Fy_ini
    sys['Gx_ini'] = Gx_ini
    sys['Gy_ini'] = Gy_ini
    
    sys['Fu_run'] = Fu_run
    sys['Gu_run'] = Gu_run

    sys['Hx_run'] = Hx_run
    sys['Hy_run'] = Hy_run
    sys['Hu_run'] = Hu_run
    
    
    Fx_ini_rows,Fx_ini_cols = np.nonzero(Fx_ini)
    Fy_ini_rows,Fy_ini_cols = np.nonzero(Fy_ini)
    Gx_ini_rows,Gx_ini_cols = np.nonzero(Gx_ini)
    Gy_ini_rows,Gy_ini_cols = np.nonzero(Gy_ini)
    
    sys['Fx_ini_rows'] = Fx_ini_rows 
    sys['Fx_ini_cols'] = Fx_ini_cols 
    sys['Fy_ini_rows'] = Fy_ini_rows 
    sys['Fy_ini_cols'] = Fy_ini_cols 
    sys['Gx_ini_rows'] = Gx_ini_rows 
    sys['Gx_ini_cols'] = Gx_ini_cols 
    sys['Gy_ini_rows'] = Gy_ini_rows 
    sys['Gy_ini_cols'] = Gy_ini_cols 

    Fx_run_rows,Fx_run_cols = np.nonzero(Fx_run)
    Fy_run_rows,Fy_run_cols = np.nonzero(Fy_run)
    Gx_run_rows,Gx_run_cols = np.nonzero(Gx_run)
    Gy_run_rows,Gy_run_cols = np.nonzero(Gy_run)
    
    sys['Fx_run_rows'] = Fx_run_rows 
    sys['Fx_run_cols'] = Fx_run_cols 
    sys['Fy_run_rows'] = Fy_run_rows 
    sys['Fy_run_cols'] = Fy_run_cols 
    sys['Gx_run_rows'] = Gx_run_rows 
    sys['Gx_run_cols'] = Gx_run_cols 
    sys['Gy_run_rows'] = Gy_run_rows 
    sys['Gy_run_cols'] = Gy_run_cols 
    

    if verbose: print(f'end system  (time: {time.time()-t_0})')   
    
    return sys


def getIndex(s, i): 
  
    # If input is invalid. 
    if s[i] != '(':
        return -1
  
    # Create a deque to use it as a stack. 
    d = deque() 
  
    # Traverse through all elements 
    # starting from i. 
    for k in range(i, len(s)): 
  
        # Pop a starting bracket 
        # for every closing bracket 
        if s[k] == ')': 
            d.popleft() 
  
        # Push all starting brackets 
        elif s[k] == '(': 
            d.append(s[i]) 
  
        # If deque becomes empty 
        if not d: 
            return k 
  
    return -1
  
def arg2np(string,function_name):
    idx_end = 0
    for it in range(3):
        if function_name in string[idx_end:]:
            #print(string[idx_end:])
            idx_ini = string.find(f'{function_name}(',idx_end)+len(f'{function_name}(')
            idx_end = getIndex(string, idx_ini-1)
            string = string.replace(string[idx_ini:idx_end],f'np.array([{string[idx_ini:idx_end]}])')
    else: pass
    return string    

def replace(eq, x_list, y_list, u_list, params_list):
    eq_str = str(eq)
    it = 0
    for item in x_list:
        x_sym = sym.Symbol(f'x_{it:04}',real=True)
        eq = eq.replace(item,x_sym)
        eq_str = str(eq)
        eq_str = eq_str.replace(str(x_sym), f'x[{it}]')
        it +=1
    it = 0
    for item in y_list:
        y_sym = sym.Symbol(f'y_{it:04}',real=True)
        eq = eq.replace(item,y_sym)
        eq_str = str(eq)
        eq_str = eq_str.replace(str(y_sym), f'y[{it}]')
        it +=1
    it = 0
    for item in u_list:
        u_sym = sym.Symbol(f'u_{it:04}')
        eq = eq.subs(str(item),u_sym)
        eq_str = str(eq)
        eq_str = eq_str.replace(str(u_sym), f'u[{it}]')
        it +=1
    it = 0
    for item in params_list:
        item_sym = sym.Symbol(item)
        p_sym = sym.Symbol(f'p_{it:04}',real=True)
        eq = eq.replace(item_sym,p_sym)
        eq_str = str(eq)
        eq_str = eq_str.replace(str(p_sym), f'p[{it}]')
        it +=1    

    eq_str = str(eq)
    
    it = 0
    for item in x_list:
        x_sym = sym.Symbol(f'x_{it:04}',real=True)        
        eq_str = eq_str.replace(str(x_sym), f'x[{it}]')
        it +=1
    it = 0
    for item in y_list:
        y_sym = sym.Symbol(f'y_{it:04}',real=True)
        eq_str = eq_str.replace(str(y_sym), f'y[{it}]')
        it +=1
    it = 0
    for item in u_list:
        u_sym = sym.Symbol(f'u_{it:04}')
        eq_str = eq_str.replace(str(u_sym), f'u[{it}]')
        it +=1
    it = 0
    for item in params_list:
        item_sym = sym.Symbol(item)
        p_sym = sym.Symbol(f'p_{it:04}',real=True)
        eq_str = eq_str.replace(str(p_sym), f'p[{it}]')
        it +=1   
        
    return eq_str

def mreplace(string,replacements):
    for item in replacements:
        string = str(string).replace(item,replacements[item])
    return string

def sym2str(sym_exp,x,y,u,p,multi_eval=False):
    
    matrix_str = str(sym_exp)
    
    mev = ''
    if multi_eval:
        mev = 'i,'
    
    if sym_exp == 0:
        matrix_str = 0
        return matrix_str
        

    for it in range(len(x)):
        name = str(x[it])
        matrix_str = re.sub(r'\b' + name + r'\b',f'x[{mev}{it}]',matrix_str)
    for it in range(len(y)):
        name = str(y[it])
        matrix_str = re.sub(r'\b' + name + r'\b',f'y[{mev}{it}]',matrix_str)
    for it in range(len(u)):
        name = str(u[it])
        matrix_str = re.sub(r'\b' + name + r'\b',f'u[{mev}{it}]',matrix_str)
    it = 0
    for item in p:
        name = str(item)
        matrix_str = re.sub(r'\b' + name + r'\b',f'p[{mev}{it}]',matrix_str)
        it+=1
        
    return matrix_str   

def vector2string(vector,function_header,vector_name,x,y,u,p,aux={},multi_eval=False):
    string = ''
    string_i = ''
    N  = len(vector)
    
    for item in aux:
        str_element = sym2str(aux[item],x,y,u,p,multi_eval=multi_eval)
        string_i += f'{" "*4}{item} = {str_element}\n'
     
    string_i += '\n'
    
    for it in range(N):
        str_element = sym2str(vector[it],x,y,u,p,multi_eval=multi_eval)
        str_element = arg2np(str_element,"Piecewise")

        if multi_eval:
            string_i += f'{" "*4}{vector_name}[i,{it}] = {str_element}\n'
        else:
            string_i += f'{" "*4}{vector_name}[{it}] = {str_element}\n'            

    string += '\n'
    string += function_header
    string += string_i
    string += '\n'
    
    return string

def matrix2string(matrix,function_header,matrix_name,x,y,u,p):
    string = ''
    string_xy = ''
    string_up = ''
    string_num = ''
    N_row,N_col = matrix.shape
    
    for irow in range(N_row):
        for icol in range(N_col):
            element = matrix[irow,icol]
            
            if element.is_number:
                str_element = str(element)
            else:
                str_element = sym2str(element,x,y,u,p)
            
            if str_element != '0':
                str_element = arg2np(str_element,"Piecewise")
                if 'x' in str_element or 'y' in str_element:
                    string_xy += f'{" "*4}{matrix_name}[{irow},{icol}] = {str_element}\n'
                elif 'u' in str_element or 'p' in str_element or 'Dt' in str_element:
                    string_up += f'{" "*4}{matrix_name}[{irow},{icol}] = {str_element}\n'
                else:
                    string_num += f'{" "*4}{matrix_name}[{irow},{icol}] = {str_element}\n'

                    

    string += '\n'
    string += '@numba.njit(cache=True)\n'
    string += function_header.replace('(','_xy(')
    string += string_xy
    string += '\n'
    string += '@numba.njit(cache=True)\n'
    string += function_header.replace('(','_up(')
    string += string_up
    string += '\n'
    string += function_header.replace('(','_num(')
    string += string_num
    string += '\n'
    
    return string

def spmatrix2string(spmatrix_list,function_header,matrix_name,x,y,u,p):
    
    data = spmatrix_list[0]
    string = ''
    string_xy = ''
    string_up = ''
    string_num = ''
    
    for irow in range(len(data)):

        element = data[irow]
        
        if element.is_number:
            str_element = str(element)
        else:
            str_element = sym2str(element,x,y,u,p)        
        
        if str_element != 0:
            str_element = arg2np(str_element,"Piecewise")
            if 'x' in str_element or 'y' in str_element:
                string_xy += f'{" "*4}{matrix_name}[{irow}] = {str_element}\n'
            elif 'u' in str_element or 'p' in str_element or 'Dt' in str_element:
                string_up += f'{" "*4}{matrix_name}[{irow}] = {str_element}\n'
            else:
                string_num+= f'{" "*4}{matrix_name}[{irow}] = {str_element}\n'
                

    string += '\n'
    string += '@numba.njit(cache=True)\n'
    string += function_header.replace('(','_xy(')
    string += string_xy
    string += '\n'
    string += '@numba.njit(cache=True)\n'
    string += function_header.replace('(','_up(')
    string += string_up
    string += '\n'
    string += function_header.replace('(','_num(')
    string += string_num
    string += '\n'
    
    string += f'def {matrix_name}_vectors():\n\n'
    string += f'{" "*4}{matrix_name}_ia = ' + str(spmatrix_list[1]) + '\n'
    string += f'{" "*4}{matrix_name}_ja = ' + str(spmatrix_list[2]) + '\n'    
    string += f'{" "*4}{matrix_name}_nia = ' + str(spmatrix_list[3][0])  + '\n'
    string += f'{" "*4}{matrix_name}_nja = ' + str(spmatrix_list[3][1])  + '\n'
    string += f'{" "*4}return {matrix_name}_ia, {matrix_name}_ja, {matrix_name}_nia, {matrix_name}_nja \n'
    
    return string


def sys2num(sys, verbose=False):
    
    params = sys['params_dict']
    h_dict = sys['h_dict']
    
    if 'xy0_dict' in sys:
        xy0_dict = sys['xy0_dict']
        
    x = sys['x']
    y_ini = sys['y_ini']
    y_run = sys['y_run']

    u_ini = sys['u_ini']
    u_run = sys['u_run']

    substitutions = {}
    replacements = {}    
    substitutions_ini = {}
    replacements_ini = {}
    substitutions_run = {}
    replacements_run = {}
    it = 0
    for item in sys['params_dict']:
        p     = sym.Symbol(f'p_{it:05}',real=True)
        param = sym.Symbol(item,real=True)
        substitutions.update({param:p})
        replacements.update({f'p_{it:05}':f'p[{it}]'})
        it+=1
    for item in sys['params_dict']:
        p     = sym.Symbol(f'p_{it:05}',real=True)
        param = sym.Symbol(item)
        substitutions.update({param:p})
        replacements.update({f'p_{it:05}':f'p[{it}]'})
        it+=1


    it = 0
    for item in sys['x']:
        x_new = sym.Symbol(f'x_{it:05}',real=True)
        param = sym.Symbol(str(item),real=True)
        substitutions.update({param:x_new})
        replacements.update({f'x_{it:05}':f'x[{it}]'})
        it+=1   
        
    
    it = 0
    for item in sys['y_ini']:
        
        new = sym.Symbol(f'y_ini_{it:05}',real=True)
        param = sym.Symbol(str(item),real=True)
        substitutions_ini.update({param:new})
        replacements_ini.update({f'y_ini_{it:05}':f'y[{it}]'})
        it+=1  
    it = 0
    for item in sys['u_ini']:
        new = sym.Symbol(f'u_ini_{it:05}',real=True)
        param = sym.Symbol(str(item),real=True)
        substitutions_ini.update({param:new})
        replacements_ini.update({f'u_ini_{it:05}':f'u[{it}]'})
        it+=1   
    it = 0
    for item in sys['y_run']:
        new = sym.Symbol(f'y_run_{it:05}',real=True)
        param = sym.Symbol(str(item),real=True)
        substitutions_run.update({param:new})
        replacements_run.update({f'y_run_{it:05}':f'y[{it}]'})
        it+=1  
    it = 0
    for item in sys['u_run']:
        new = sym.Symbol(f'u_run_{it:05}',real=True)
        param = sym.Symbol(str(item),real=True)
        substitutions_run.update({param:new})
        replacements_run.update({f'u_run_{it:05}':f'u[{it}]'})
        it+=1  
        
    
        
    
    f_ini = sys['f']
    g_ini = sys['g'] 
    
    f_run = sys['f']
    g_run = sys['g'] 
    
    h = sys['h'] 

    Fx_run = sys['Fx_run']
    Fy_run = sys['Fy_run']
    Gx_run = sys['Gx_run']
    Gy_run = sys['Gy_run']
    
    Fx_ini = sys['Fx_run']
    Fy_ini = sys['Fy_ini']
    Gx_ini = sys['Gx_ini']
    Gy_ini = sys['Gy_ini']

    Fu_run = sys['Fu_run']
    Gu_run = sys['Gu_run'] 
    
    Hx_run = sys['Hx_run'] 
    Hy_run = sys['Hy_run'] 
    Hu_run = sys['Hu_run']

    
    N_x = len(x)
    N_y = len(y_run)
    N_u = len(u_run)
    N_z = len(h_dict)
    
    t_0 = time.time()
    if verbose: print(f'computing jac_ini (time: {time.time()-t_0})')
    jac_ini = sym.Matrix([[Fx_ini,Fy_ini],[Gx_ini,Gy_ini]]) 
    if verbose: print(f'computing jac_run (time: {time.time()-t_0})')
    jac_run = sym.Matrix([[Fx_run,Fy_run],[Gx_run,Gy_run]]) 
    
    if verbose: print(f'computing jac_trap (time: {time.time()-t_0})')
    eye = sym.eye(N_x, real=True)
    Dt = sym.Symbol('Dt',real=True)
    jac_trap = sym.Matrix([[eye - 0.5*Dt*Fx_run, -0.5*Dt*Fy_run],[Gx_run,Gy_run]])    

    sys['jac_ini']  = jac_ini
    sys['jac_run']  = jac_run   
    sys['jac_trap'] = jac_trap
    

    name = sys['name']
    
    if 'aux_dict' in sys:
        aux_dict = sys['aux_dict']
    else:
        aux_dict = {}
        

    f_ini_eval = ''
    f_run_eval = ''
    
    
    numba_enable = True
    tab = ' '*4

    if verbose: print(f'generating f_ini_eval string (time: {time.time()-t_0})')    
    # f: differential equations for backward problem
    function_header  = '@numba.njit(cache=True)\n'
    function_header += 'def f_ini_eval(f_ini,x,y,u,p,xyup = 0):\n\n'
    x,y,u,p = sys['x'],sys['y_ini'],sys['u_ini'],sys['params_dict'].keys()
    f_ini_eval += vector2string(sys['f'],function_header,'f_ini',x,y,u,p,aux=aux_dict,multi_eval=False)

    if verbose: print(f'generating f_run_eval string (time: {time.time()-t_0})')        
    # f: differential equations for foreward problem
    function_header  = '@numba.njit(cache=True)\n'
    function_header += 'def f_run_eval(f_run,x,y,u,p,xyup = 0):\n\n'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    f_run_eval += vector2string(sys['f'],function_header,'f_run',x,y,u,p,aux=aux_dict,multi_eval=False)


    if verbose: print(f'generating g_ini_eval string (time: {time.time()-t_0})')        
    # g: algebraic equations for backward problem
    function_header  = '@numba.njit(cache=True)\n'
    function_header += 'def g_ini_eval(g_ini,x,y,u,p,xyup = 0):\n\n'
    x,y,u,p = sys['x'],sys['y_ini'],sys['u_ini'],sys['params_dict'].keys()
    g_ini_eval = vector2string(sys['g'],function_header,'g_ini',x,y,u,p,aux=aux_dict,multi_eval=False)

    if verbose: print(f'generating g_run_eval string (time: {time.time()-t_0})')           
    # g: algebraic equations for forward problem
    function_header  = '@numba.njit(cache=True)\n'
    function_header += 'def g_run_eval(g_run,x,y,u,p,xyup = 1):\n\n'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    g_run_eval = vector2string(sys['g'],function_header,'g_run',x,y,u,p,aux=aux_dict,multi_eval=False)

    if verbose: print(f'generating h_run_eval string (time: {time.time()-t_0})')               
    # h: outputs for the foreward problem
    function_header  = '@numba.njit(cache=True)\n'
    function_header += 'def h_eval(h_run,x,y,u,p,xyup = 1):\n\n'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    h_run_eval = vector2string(sys['h'],function_header,'h_run',x,y,u,p,aux=aux_dict,multi_eval=False)

    if verbose: print(f'generating f_run_gpu string (time: {time.time()-t_0})')      
    # GPU f: differential equations for foreward problem
    function_header  = '@cuda.jit(device=True)\n'
    function_header += 'def f_run_gpu(f_run,x,u,p):\n\n'
    function_header += '    sin = math.sin\n'
    function_header += '    cos = math.cos\n'
    function_header += '    sqrt = math.sqrt\n'
    function_header += '    abs = math.fabs\n'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    f_run_gpu = vector2string(sys['f'],function_header,'f_run',x,y,u,p,aux=aux_dict,multi_eval=False)

    if verbose: print(f'generating h_run_gpu string (time: {time.time()-t_0})')      
    # GPU h: outputs
    function_header  = '@cuda.jit(device=True)\n'
    function_header += 'def h_eval_gpu(z,x,u,p):\n\n'
    function_header += '    sin = math.sin\n'
    function_header += '    cos = math.cos\n'
    function_header += '    sqrt = math.sqrt\n'
    function_header += '    abs = math.fabs\n'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    h_run_gpu = vector2string(sys['h'],function_header,'z',x,y,u,p,aux=aux_dict,multi_eval=False)



  
    xy0_eval = ''
    if 'xy0_dict' in sys:    
        if numba_enable: xy0_eval += '@numba.njit(cache=True)\n'
        xy0_eval += 'def xy0_eval(x,y,u,p):\n\n'
        for item in xy0_dict:
            if type(xy0_dict[item]) == float:
                xy0_sym_ = sym.Symbol(str(xy0_dict[item]))
            else:
                xy0_sym_ = xy0_dict[item]
            x,y,u,p = sys['x'],sys['y_ini'],sys['u_ini'],sys['params_dict'].keys()
            rh = sym2str(xy0_sym_,x,y,u,p,multi_eval=False)
            item_sym = sym.Symbol(item)
            lh = sym2str(item_sym,x,y,u,p,multi_eval=False)
            xy0_eval += f'{tab}{lh} = {arg2np(rh,"Piecewise")}\n'
        xy0_eval += f'{tab}\n'    
        
    
### jacobian steady state backward 
    if verbose: print(f'generating jac_ini_ss_eval string (time: {time.time()-t_0})')       
    function_header = 'def jac_ini_ss_eval(jac_ini,x,y,u,p,xyup = 1):\n\n'
    matrix_name = 'jac_ini'
    x,y,u,p = sys['x'],sys['y_ini'],sys['u_ini'],sys['params_dict'].keys()
    jac_ini_ss_eval = matrix2string(jac_ini,function_header,matrix_name,x,y,u,p)

    
## sparse jacobian ini 
    if verbose: print(f'generating sp_jac_ini_eval string (time: {time.time()-t_0})')      
    spmatrix_list = _doktocsr(SparseMatrix(jac_ini))
    function_header =  '@numba.njit(cache=True)\n'
    function_header = 'def sp_jac_ini_eval(sp_jac_ini,x,y,u,p,Dt,xyup = 1):\n\n'
    matrix_name = 'sp_jac_ini'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    sp_jac_ini_eval = spmatrix2string(spmatrix_list,function_header,matrix_name,x,y,u,p)
    
    
    
## jacobian steady state forward
    if verbose: print(f'generating jac_ss_eval string (time: {time.time()-t_0})')      
    function_header =  '@numba.njit(cache=True)\n'
    function_header = 'def jac_run_ss_eval(jac_run,x,y,u,p,xyup = 1):\n\n'
    matrix_name = 'jac_run'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    jac_ss_eval = matrix2string(jac_run,function_header,matrix_name,x,y,u,p)


## jacobian trapezoidal 
    if verbose: print(f'generating jac_trap_eval string (time: {time.time()-t_0})')      
    function_header =  '@numba.njit(cache=True)\n'
    function_header = 'def jac_trap_eval(jac_trap,x,y,u,p,Dt,xyup = 1):\n\n'
    matrix_name = 'jac_trap'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    jac_trap_eval = matrix2string(jac_trap,function_header,matrix_name,x,y,u,p)

## sparse jacobian trapezoidal 
    if verbose: print(f'generating sp_jac_trap_eval string (time: {time.time()-t_0})')      
    spmatrix_list = _doktocsr(SparseMatrix(jac_trap))
    function_header =  '@numba.njit(cache=True)\n'
    function_header = 'def sp_jac_trap_eval(sp_jac_trap,x,y,u,p,Dt,xyup = 1):\n\n'
    matrix_name = 'sp_jac_trap'
    x,y,u,p = sys['x'],sys['y_run'],sys['u_run'],sys['params_dict'].keys()
    sp_jac_trap_eval = spmatrix2string(spmatrix_list,function_header,matrix_name,x,y,u,p)
    


    class_template = pkgutil.get_data(__name__, "templates/class_dae_fast_template.py").decode().replace('\r\n','\n') 
    functions_template = pkgutil.get_data(__name__, "templates/functions_template.py").decode().replace('\r\n','\n') 
#     solver_template = pkgutil.get_data(__name__, "templates/solver_template_v2.py").decode().replace('\r\n','\n') 

    class_template = class_template.replace('{name}',str(name))
    class_template = class_template.replace('{N_x}',str(N_x)).replace('{N_y}',str(N_y)).replace('{N_z}',str(N_z))
    class_template = class_template.replace('{x_list}', str([str(item) for item in x]))
    class_template = class_template.replace('{y_run_list}', str([str(item) for item in y_run]))
    class_template = class_template.replace('{y_ini_list}', str([str(item) for item in y_ini]))
    class_template = class_template.replace('{params_list}', str([str(item) for item in params]))
    class_template = class_template.replace('{params_values_list}', str([(params[item]) for item in params]))
    class_template = class_template.replace('{inputs_ini_list}', str([str(item) for item in u_ini]))
    class_template = class_template.replace('{inputs_run_list}', str([str(item) for item in u_run]))
    class_template = class_template.replace('{inputs_ini_values_list}', str([(sys['u_ini_dict'][item]) for item in sys['u_ini_dict']]))
    class_template = class_template.replace('{inputs_run_values_list}', str([(sys['u_run_dict'][item]) for item in sys['u_run_dict']]))
    class_template = class_template.replace('{outputs_list}', str([str(item) for item in h_dict]))

    if verbose: print(f'writting file (time: {time.time()-t_0}')      
    module = class_template
    module += '\n'*3
    module += f_ini_eval
    module += '\n'*3
    module += f_run_eval
    module += '\n'*3
    module += g_ini_eval
    module += '\n'*3
    module += g_run_eval
    module += '\n'*3    
    module += f_run_gpu
    module += '\n'*3
    module += h_run_gpu
    module += '\n'*3  

    module += jac_ini_ss_eval
    module += '\n'*3
    module += sp_jac_ini_eval
    module += '\n'*3
    module += jac_ss_eval
    module += '\n'*3
    module += jac_trap_eval
    module += '\n'*3
    module += h_run_eval
    module += '\n'*3
    module += functions_template
    module += '\n'*3
    module += xy0_eval
    module += '\n'*3
    module += sp_jac_trap_eval
    module += '\n'*3    
    module += 'def nonzeros():\n'
    module += f"    Fx_ini_rows = {sys['Fx_ini_rows'].tolist()}\n\n"
    module += f"    Fx_ini_cols = {sys['Fx_ini_cols'].tolist()}\n\n"   
    module += f"    Fy_ini_rows = {sys['Fy_ini_rows'].tolist()}\n\n"
    module += f"    Fy_ini_cols = {sys['Fy_ini_cols'].tolist()}\n\n"  
    module += f"    Gx_ini_rows = {sys['Gx_ini_rows'].tolist()}\n\n"
    module += f"    Gx_ini_cols = {sys['Gx_ini_cols'].tolist()}\n\n"  
    module += f"    Gy_ini_rows = {sys['Gy_ini_rows'].tolist()}\n\n"
    module += f"    Gy_ini_cols = {sys['Gy_ini_cols'].tolist()}\n\n"    
    module += f"    return Fx_ini_rows,Fx_ini_cols,Fy_ini_rows,Fy_ini_cols,Gx_ini_rows,Gx_ini_cols,Gy_ini_rows,Gy_ini_cols"       

    
    with open(f'{name}.py','w') as fobj:
        fobj.write(module)
   
    if verbose: print(f'sys2num (time: {time.time()-t_0})')      





    
    
if __name__ == "__main__":

    
    params_dict = {'X_d':1.81,'X1d':0.3, 'T1d0':8.0,  # synnchronous machine d-axis parameters
                   'X_q':1.76,'X1q':0.65,'T1q0':1.0,  # synnchronous machine q-axis parameters
                   'R_a':0.003,
                   'X_l': 0.02, 
                   'H':3.5,'D':0.0,
                   'Omega_b':2*np.pi*50,'omega_s':1.0,
                   'V_0':1.0,'Theta_0':0.0}
    
    
    u_ini_dict = {'p_m':0.8,'v_f':1.0}  # for the initialization problem
    u_run_dict = {'p_m':0.8,'v_f':1.0}  # for the running problem (here initialization and running problem are the same)
    
    
    x_list = ['delta','omega','e1q','e1d']    # dynamic states
    y_ini_list = ['i_d','i_q','p_t','q_t','v_t','theta_t']   
    y_run_list = ['i_d','i_q','p_t','q_t','v_t','theta_t']   
    
    sys_vars = {'params':params_dict,
                'u_list':u_run_dict,
                'x_list':x_list,
                'y_list':y_run_list}
    
    exec(sym_gen_str())  # exec to generate the required symbolic varables and constants

    # auxiliar equations
    v_d = v_t*sin(delta - theta_t)  # park
    v_q = v_t*cos(delta - theta_t)  # park
    
    p_e = i_d*(v_d + R_a*i_d) + i_q*(v_q + R_a*i_q) # electromagnetic power
    
    # dynamic equations
    ddelta = Omega_b*(omega - omega_s) - 1e-4*delta  # load angle
    domega = 1/(2*H)*(p_m - p_e - D*(omega - omega_s)) # speed
    de1q = 1/T1d0*(-e1q - (X_d - X1d)*i_d + v_f)
    de1d = 1/T1q0*(-e1d + (X_q - X1q)*i_q)
    
    # algrbraic equations
    g_1 = v_q + R_a*i_q + X1d*i_d - e1q # stator
    g_2 = v_d + R_a*i_d - X1q*i_q - e1d # stator
    g_3 = i_d*v_d + i_q*v_q - p_t # active power 
    g_4 = i_d*v_q - i_q*v_d - q_t # reactive power
    g_5 = p_t - (v_t*V_0*sin(theta_t - Theta_0))/X_l  # network equation (p)
    g_6 = q_t + (v_t*V_0*cos(theta_t - Theta_0))/X_l - v_t**2/X_l  # network equation (q) 

    sys_dict = {'name':'build_test',
           'params_dict':params,
           'f_list':[ddelta,domega,de1q,de1d],
           'g_list':[g_1,g_2,g_3,g_4,g_5,g_6],
           'x_list':x_list,
           'y_ini_list':y_ini_list,
           'y_run_list':y_run_list,
           'u_run_dict':u_run_dict,
           'u_ini_dict':u_ini_dict,
           'h_dict':{'p_m':p_m,'p_e':p_e, 'v_f':v_f}}
    
    sys_dict = system(sys_dict)
   # sys2num(sys_dict)                                                             



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020

@author: jmmauricio
"""

import numpy as np
from cycler import cycler


def set_style(plt):

    plt.rcParams['axes.prop_cycle'] = cycler('color', ['#d9524f', '#5cb85c', '#337ab7', '#f0ad4e', '#5bc0de','#5e4485'])
    #cycler('color',['#d9524f', '5cb85c', '337ab7', 'f0ad4e', '5bc0de','5e4485']) # jmcolo 
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    # Set default figure size
    plt.rcParams['figure.figsize'] = (3.3, 2.5)
    #plt.rcParams['figure.dpi'] : 600

    # Font sizes
    plt.rcParams['font.size'] =  10
    plt.rcParams['font.family'] = 'serif'
    plt.rcParams['font.serif'] = 'Times New Roman'
    plt.rcParams['mathtext.fontset'] = 'cm' #'stix'
    #plt.rcParams['text.usetex'] = True
    plt.rcParams['legend.handlelength'] = 1.

    return colors


